# pylint: disable = too-many-lines

from dataclasses import dataclass, field as dataclass_field
from typing import Callable, Iterable, List, NoReturn, Sequence, Set, Tuple

from rflx import (
    declaration as decl,
    expression as expr,
    identifier as rid,
    model,
    statement as stmt,
    typing_ as rty,
)
from rflx.ada import (
    FALSE,
    ID,
    TRUE,
    Add,
    And,
    AndThen,
    Assignment,
    Call,
    CallStatement,
    CaseStatement,
    ContextItem,
    Declaration,
    Declare,
    EnumerationType,
    Equal,
    ExitStatement,
    Expr,
    ExpressionFunctionDeclaration,
    First,
    FormalDeclaration,
    FormalSubprogramDeclaration,
    FunctionSpecification,
    GenericProcedureInstantiation,
    GreaterEqual,
    IfStatement,
    Indexed,
    Last,
    Length,
    Less,
    LessEqual,
    Mod,
    Mul,
    NamedAggregate,
    New,
    Not,
    NotEqual,
    NullStatement,
    Number,
    ObjectDeclaration,
    Or,
    OutParameter,
    Parameter,
    Postcondition,
    Pragma,
    PragmaStatement,
    Precondition,
    ProcedureSpecification,
    QualifiedExpr,
    ReturnStatement,
    Size,
    SizeAspect,
    Slice,
    Statement,
    String,
    SubprogramBody,
    SubprogramDeclaration,
    UnitPart,
    UseTypeClause,
    ValueRange,
    Variable,
    While,
    WithClause,
)
from rflx.const import BUILTINS_PACKAGE, INTERNAL_PACKAGE
from rflx.error import Subsystem, fail, fatal_fail

from . import const


@dataclass
class SessionContext:
    referenced_types: Set[rid.ID] = dataclass_field(default_factory=set)
    referenced_types_body: Set[rid.ID] = dataclass_field(default_factory=set)
    referenced_packages_body: Set[rid.ID] = dataclass_field(default_factory=set)
    referenced_has_data: Set[rid.ID] = dataclass_field(default_factory=set)
    used_types_body: Set[rid.ID] = dataclass_field(default_factory=set)
    state_exception: Set[rid.ID] = dataclass_field(default_factory=set)


@dataclass
class EvaluatedDeclaration:
    global_declarations: List[Declaration] = dataclass_field(default_factory=list)
    initialization_declarations: List[Declaration] = dataclass_field(default_factory=list)
    initialization: List[Statement] = dataclass_field(default_factory=list)
    finalization: List[Statement] = dataclass_field(default_factory=list)

    def __iadd__(self, other: object) -> "EvaluatedDeclaration":
        if isinstance(other, EvaluatedDeclaration):
            return EvaluatedDeclaration(
                [*self.global_declarations, *other.global_declarations],
                [*self.initialization_declarations, *other.initialization_declarations],
                [*self.initialization, *other.initialization],
                [*self.finalization, *other.finalization],
            )

        return NotImplemented


@dataclass(frozen=True)
class ExceptionHandler:
    state: model.State
    finalization: Sequence[Statement]

    def execute(self) -> Sequence[Statement]:
        assert (
            self.state.exception_transition
        ), f'missing exception transition for state "{self.state.identifier}"'
        return [
            Assignment(
                "State",
                Variable(f"S_{self.state.exception_transition.target}"),
            ),
            *self.finalization,
            ReturnStatement(),
        ]


class SessionGenerator:  # pylint: disable = too-many-instance-attributes
    def __init__(self, session: model.Session, prefix: str = "", debug: bool = False) -> None:
        self._session = session
        self._prefix = prefix
        self._debug = debug

        self._session_context = SessionContext()
        self._declaration_context: List[ContextItem] = []
        self._body_context: List[ContextItem] = []
        self._formal_parameters: List[FormalDeclaration] = []
        self._unit_part = UnitPart()

        self._create()

    @property
    def unit_identifier(self) -> ID:
        return ID(self._session.identifier)

    @property
    def declaration_context(self) -> List[ContextItem]:
        return self._declaration_context

    @property
    def body_context(self) -> List[ContextItem]:
        return self._body_context

    @property
    def formal_parameters(self) -> List[FormalDeclaration]:
        return self._formal_parameters

    @property
    def unit_part(self) -> UnitPart:
        return self._unit_part

    def _ada_type(self, identifier: rid.ID) -> ID:
        if model.is_builtin_type(identifier):
            return ID(identifier.name)

        return ID(model.qualified_type_identifier(identifier, self._session.package))

    def _model_type(self, identifier: rid.ID) -> model.Type:
        return self._session.types[
            model.qualified_type_identifier(identifier, self._session.package)
        ]

    def _create(self) -> None:
        self._unit_part = self._create_state_machine()

        for parameter in self._session.parameters.values():
            if isinstance(parameter, decl.ChannelDeclaration):
                if (
                    parameter.readable
                    and parameter.identifier not in self._session_context.referenced_has_data
                ):
                    self._unit_part.specification.insert(
                        0,
                        Pragma("Unreferenced", [Variable(ID(parameter.identifier + "_Has_Data"))]),
                    )

        self._formal_parameters = self._create_formal_parameters(self._session.parameters.values())
        self._declaration_context, self._body_context = self._create_context()

    def _create_formal_parameters(
        self,
        parameters: Iterable[decl.FormalDeclaration],
    ) -> List[FormalDeclaration]:
        result: List[FormalDeclaration] = []

        for parameter in parameters:
            if isinstance(parameter, decl.ChannelDeclaration):
                result.extend(self._create_channel_parameters(parameter))
            elif isinstance(parameter, decl.FunctionDeclaration):
                result.extend(self._create_function_parameters(parameter))
            elif isinstance(parameter, decl.TypeDeclaration):
                fail(
                    f'type declaration "{parameter.identifier}" not yet supported',
                    Subsystem.GENERATOR,
                    location=parameter.location,
                )
            else:
                fatal_fail(
                    f'unexpected formal parameter "{parameter.identifier}"',
                    Subsystem.GENERATOR,
                    location=parameter.location,
                )

        return result

    @staticmethod
    def _create_channel_parameters(channel: decl.ChannelDeclaration) -> List[FormalDeclaration]:
        return [
            *(
                [
                    FormalSubprogramDeclaration(
                        FunctionSpecification(
                            ID(f"{channel.identifier}_Has_Data"),
                            "Boolean",
                        )
                    ),
                    FormalSubprogramDeclaration(
                        ProcedureSpecification(
                            ID(f"{channel.identifier}_Read"),
                            [
                                OutParameter(["Buffer"], const.TYPES_BYTES),
                                OutParameter(["Length"], const.TYPES_LENGTH),
                            ],
                        )
                    ),
                ]
                if channel.readable
                else []
            ),
            *(
                [
                    FormalSubprogramDeclaration(
                        ProcedureSpecification(
                            ID(f"{channel.identifier}_Write"),
                            [
                                Parameter(["Buffer"], const.TYPES_BYTES),
                            ],
                        )
                    )
                ]
                if channel.writable
                else []
            ),
        ]

    def _create_function_parameters(
        self, function: decl.FunctionDeclaration
    ) -> List[FormalDeclaration]:
        procedure_parameters: List[Parameter] = []

        if function.type_ == rty.Undefined():
            fatal_fail(
                f'return type of function "{function.identifier}" is undefined',
                Subsystem.GENERATOR,
                location=function.location,
            )
        if function.type_ == rty.OPAQUE:
            fatal_fail(
                f'Opaque as return type of function "{function.identifier}" not allowed',
                Subsystem.GENERATOR,
                location=function.location,
            )
        if isinstance(function.type_, rty.Sequence):
            fail(
                f'sequence as return type of function "{function.identifier}" not yet supported',
                Subsystem.GENERATOR,
                location=function.location,
            )
        if isinstance(function.type_, rty.Message):
            if not function.type_.is_definite:
                fatal_fail(
                    "non-definite message"
                    f' in return type of function "{function.identifier}" not allowed',
                    Subsystem.GENERATOR,
                    location=function.location,
                )
            if any(
                isinstance(field_type, rty.Sequence) and not field_type == rty.OPAQUE
                for field_type in function.type_.field_types.values()
            ):
                fail(
                    "message containing sequence fields"
                    f' in return type of function "{function.identifier}" not yet supported',
                    Subsystem.GENERATOR,
                    location=function.location,
                )

        procedure_parameters.append(
            OutParameter(
                [ID(function.identifier)],
                ID(self._prefix * function.return_type * "Structure")
                if isinstance(function.type_, rty.Message)
                else ID(self._prefix * function.return_type),
            )
        )

        self._session_context.referenced_types.add(function.return_type)

        for a in function.arguments:
            if isinstance(a.type_, rty.Sequence) and not a.type_ == rty.OPAQUE:
                fail(
                    f'sequence as parameter of function "{function.identifier}" not yet supported',
                    Subsystem.GENERATOR,
                    location=function.location,
                )
            procedure_parameters.append(
                Parameter(
                    [ID(a.identifier)],
                    ID(const.TYPES_BYTES)
                    if a.type_ == rty.OPAQUE
                    else ID(self._prefix * a.type_identifier * "Structure")
                    if isinstance(a.type_, rty.Message)
                    else ID(self._prefix * a.type_identifier),
                )
            )

            assert isinstance(a.type_, (rty.Integer, rty.Enumeration, rty.Message, rty.Sequence))

            self._session_context.referenced_types.add(a.type_.identifier)

        return [
            FormalSubprogramDeclaration(
                ProcedureSpecification(
                    ID(function.identifier),
                    procedure_parameters,
                )
            )
        ]

    def _create_context(self) -> Tuple[List[ContextItem], List[ContextItem]]:
        declaration_context: List[ContextItem] = [
            WithClause(self._prefix * const.TYPES_PACKAGE),
        ]
        body_context: List[ContextItem] = [
            *([WithClause("Ada.Text_IO")] if self._debug else []),
        ]

        for referenced_types, context in [
            (self._session_context.referenced_types, declaration_context),
            (self._session_context.referenced_types_body, body_context),
        ]:
            for type_identifier in referenced_types:
                if type_identifier.parent in [INTERNAL_PACKAGE, BUILTINS_PACKAGE]:
                    continue
                type_ = self._model_type(type_identifier)
                context.extend(
                    [
                        WithClause(self._prefix * ID(type_.package)),
                        *(
                            [
                                WithClause(self._prefix * ID(type_.identifier)),
                            ]
                            if isinstance(type_, (model.Message, model.Sequence))
                            else []
                        ),
                    ]
                )

        body_context.extend(
            [
                WithClause(self._prefix * ID(p))
                for p in self._session_context.referenced_packages_body
            ],
        )

        for type_identifier in self._session_context.used_types_body:
            if type_identifier.parent in [INTERNAL_PACKAGE, BUILTINS_PACKAGE]:
                continue
            if type_identifier in [const.TYPES_LENGTH, const.TYPES_INDEX, const.TYPES_BIT_LENGTH]:
                body_context.append(
                    WithClause(self._prefix * const.TYPES_PACKAGE),
                )
            body_context.append(
                UseTypeClause(self._prefix * ID(type_identifier)),
            )

        use_type_clauses = {i for i in body_context if isinstance(i, UseTypeClause)}
        body_context = [
            i
            for i in body_context
            if isinstance(i, UseTypeClause)
            or (
                isinstance(i, WithClause)
                and (
                    i not in declaration_context
                    or any(utc.identifier.parent == i.identifier for utc in use_type_clauses)
                )
            )
        ]

        return (declaration_context, body_context)

    def _create_state_machine(self) -> UnitPart:
        evaluated_declarations = self._evaluate_declarations(
            self._session.declarations.values(),
            session_global=True,
        )

        unit = UnitPart()
        unit += self._create_declarations(self._session, evaluated_declarations.global_declarations)
        unit += self._create_uninitialized_function(self._session.declarations.values())
        unit += self._create_states(self._session)
        unit += self._create_initialized_procedure(self._session)
        unit += self._create_active_function(self._session)
        unit += self._create_initialize_procedure(
            self._session,
            evaluated_declarations.initialization_declarations,
            evaluated_declarations.initialization,
        )
        unit += self._create_finalize_procedure(
            self._session,
            evaluated_declarations.initialization_declarations,
            evaluated_declarations.finalization,
        )
        unit += self._create_tick_procedure(self._session)
        unit += self._create_run_procedure()
        return unit

    @staticmethod
    def _create_declarations(
        session: model.Session, declarations: Sequence[Declaration]
    ) -> UnitPart:
        return UnitPart(
            private=[
                UseTypeClause(const.TYPES_INDEX),
                EnumerationType(
                    "Session_State", {ID(f"S_{s.identifier}"): None for s in session.states}
                ),
                ObjectDeclaration(["State"], "Session_State", Variable(f"S_{session.initial}")),
                *declarations,
            ],
        )

    @staticmethod
    def _create_uninitialized_function(declarations: Iterable[decl.BasicDeclaration]) -> UnitPart:
        specification = FunctionSpecification("Uninitialized", "Boolean")
        return UnitPart(
            [
                SubprogramDeclaration(specification),
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    And(
                        *[
                            Not(
                                Call(
                                    ID(declaration.type_.identifier * "Has_Buffer"),
                                    [Variable(context_id(declaration.identifier))],
                                )
                            )
                            for declaration in declarations
                            if isinstance(declaration, decl.VariableDeclaration)
                            and isinstance(declaration.type_, (rty.Message, rty.Sequence))
                        ]
                    ),
                ),
            ],
        )

    def _create_states(self, session: model.Session) -> UnitPart:
        unit_body: List[Declaration] = []

        for state in session.states:
            if not state.is_null:
                evaluated_declarations = self._evaluate_declarations(state.declarations.values())
                actions = [
                    s
                    for a in state.actions
                    for s in self._state_action(
                        a, ExceptionHandler(state, evaluated_declarations.finalization)
                    )
                ]
                if state.identifier in self._session_context.state_exception:
                    evaluated_declarations.global_declarations.append(
                        ObjectDeclaration(["RFLX_Exception"], "Boolean", FALSE)
                    )

                unit_body += [
                    SubprogramBody(
                        ProcedureSpecification(
                            ID(state.identifier), [OutParameter(["State"], "Session_State")]
                        ),
                        [
                            *evaluated_declarations.global_declarations,
                            *evaluated_declarations.initialization_declarations,
                        ],
                        [
                            *evaluated_declarations.initialization,
                            *actions,
                            *(
                                [
                                    IfStatement(
                                        [
                                            (
                                                t.condition.substituted(
                                                    self._substitution()
                                                ).ada_expr(),
                                                [Assignment("State", Variable(f"S_{t.target}"))],
                                            )
                                            for t in state.transitions[:-1]
                                        ],
                                        [
                                            Assignment(
                                                "State",
                                                Variable(f"S_{state.transitions[-1].target}"),
                                            )
                                        ],
                                    )
                                ]
                                if state.transitions
                                else []
                            ),
                            *evaluated_declarations.finalization,
                        ],
                        aspects=[
                            Precondition(Call("Initialized")),
                            Postcondition(Call("Initialized")),
                        ],
                    )
                ]

        return UnitPart(body=unit_body)

    @staticmethod
    def _create_initialized_procedure(session: model.Session) -> UnitPart:
        specification = FunctionSpecification("Initialized", "Boolean")
        return UnitPart(
            [
                SubprogramDeclaration(specification),
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    AndThen(
                        *[
                            e
                            for d in session.declarations.values()
                            if isinstance(d, decl.VariableDeclaration)
                            and isinstance(d.type_, (rty.Message, rty.Sequence))
                            for e in [
                                Call(
                                    ID(d.type_identifier) * "Has_Buffer",
                                    [Variable(context_id(d.identifier))],
                                ),
                                Equal(
                                    Variable(context_id(d.identifier) * "Buffer_First"),
                                    First(const.TYPES_INDEX),
                                ),
                                Equal(  # ISSUE: Componolit/RecordFlux#713
                                    Variable(context_id(d.identifier) * "Buffer_Last"),
                                    Add(First(const.TYPES_INDEX), Number(4095)),
                                ),
                            ]
                        ]
                    ),
                ),
            ],
        )

    @staticmethod
    def _create_active_function(session: model.Session) -> UnitPart:
        specification = FunctionSpecification("Active", "Boolean")
        return UnitPart(
            [
                SubprogramDeclaration(specification),
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    NotEqual(Variable("State"), Variable(f"S_{session.final}")),
                ),
            ],
        )

    @staticmethod
    def _create_initialize_procedure(
        session: model.Session,
        declarations: Sequence[Declaration],
        initialization: Sequence[Statement],
    ) -> UnitPart:
        specification = ProcedureSpecification("Initialize")
        return UnitPart(
            [
                SubprogramDeclaration(
                    specification, [Postcondition(And(Call("Initialized"), Call("Active")))]
                ),
            ],
            [
                SubprogramBody(
                    specification,
                    declarations,
                    [
                        *initialization,
                        Assignment("State", Variable(f"S_{session.initial}")),
                    ],
                ),
            ],
        )

    @staticmethod
    def _create_finalize_procedure(
        session: model.Session,
        declarations: Sequence[Declaration],
        finalization: Sequence[Statement],
    ) -> UnitPart:
        specification = ProcedureSpecification("Finalize")
        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Precondition(Call("Initialized")),
                        Postcondition(And(Call("Uninitialized"), Not(Call("Active")))),
                    ],
                ),
            ],
            [
                SubprogramBody(
                    specification,
                    declarations,
                    [
                        *finalization,
                        Assignment("State", Variable(f"S_{session.final}")),
                    ],
                ),
            ],
        )

    def _create_tick_procedure(
        self,
        session: model.Session,
    ) -> UnitPart:
        specification = ProcedureSpecification("Tick")
        return UnitPart(
            [
                Pragma("Warnings", [Variable("Off"), String('subprogram "Tick" has no effect')]),
                SubprogramDeclaration(
                    specification,
                    [Precondition(Variable("Initialized")), Postcondition(Variable("Initialized"))],
                ),
                Pragma("Warnings", [Variable("On"), String('subprogram "Tick" has no effect')]),
            ],
            [
                SubprogramBody(
                    specification,
                    [],
                    [
                        CaseStatement(
                            Variable("State"),
                            [
                                (
                                    Variable(f"S_{s.identifier}"),
                                    [
                                        *self._debug_output(f"State: {s.identifier}"),
                                        CallStatement(ID(s.identifier), [Variable("State")])
                                        if s.identifier != session.final
                                        else NullStatement(),
                                    ],
                                )
                                for s in session.states
                            ],
                        ),
                    ],
                )
            ],
        )

    @staticmethod
    def _create_run_procedure() -> UnitPart:
        specification = ProcedureSpecification("Run")
        return UnitPart(
            [
                Pragma("Warnings", [Variable("Off"), String('subprogram "Run" has no effect')]),
                SubprogramDeclaration(
                    specification,
                    [
                        Postcondition(Call("Uninitialized")),
                    ],
                ),
                Pragma("Warnings", [Variable("On"), String('subprogram "Run" has no effect')]),
            ],
            [
                SubprogramBody(
                    specification,
                    [],
                    [
                        CallStatement("Initialize"),
                        While(
                            Call("Active"),
                            [
                                PragmaStatement("Loop_Invariant", [Variable("Initialized")]),
                                CallStatement("Tick"),
                            ],
                        ),
                        CallStatement("Finalize"),
                    ],
                )
            ],
        )

    def _evaluate_declarations(
        self,
        declarations: Iterable[decl.BasicDeclaration],
        session_global: bool = False,
    ) -> EvaluatedDeclaration:
        result = EvaluatedDeclaration()

        for declaration in declarations:
            if isinstance(declaration, decl.VariableDeclaration):
                result += self._declare(
                    declaration.identifier,
                    declaration.type_,
                    declaration.expression,
                    session_global=session_global,
                )
            elif isinstance(declaration, decl.RenamingDeclaration):
                fail(
                    f'renaming declaration "{declaration.identifier}" not yet supported',
                    Subsystem.GENERATOR,
                    location=declaration.location,
                )
            else:
                fatal_fail(
                    f'unexpected declaration "{declaration.identifier}"',
                    Subsystem.GENERATOR,
                    location=declaration.location,
                )

        return result

    def _state_action(
        self, action: stmt.Statement, exception_handler: ExceptionHandler
    ) -> Sequence[Statement]:
        if isinstance(action, stmt.Assignment):
            return self._assign(
                action.identifier, action.type_, action.expression, exception_handler
            )

        if isinstance(action, stmt.Append):
            return self._append(action, exception_handler)

        if isinstance(action, stmt.Extend):
            fail(
                "Extend statement not yet supported",
                Subsystem.GENERATOR,
                location=action.location,
            )

        if isinstance(action, stmt.Reset):
            return self._reset(action)

        if isinstance(action, stmt.Read):
            return self._read(action)

        if isinstance(action, stmt.Write):
            return self._write(action, exception_handler)

        fatal_fail(
            f'unexpected statement "{type(action).__name__}"',
            Subsystem.GENERATOR,
            location=action.location,
        )

    def _declare(
        self,
        identifier: rid.ID,
        type_: rty.Type,
        expression: expr.Expr = None,
        constant: bool = False,
        session_global: bool = False,
    ) -> EvaluatedDeclaration:
        result = EvaluatedDeclaration()

        if type_ == rty.OPAQUE:
            initialization = expression
            object_type: Expr = Variable(const.TYPES_BYTES)
            aspects = []

            if isinstance(expression, expr.Aggregate):
                if len(expression.elements) == 0:
                    object_type = Slice(
                        Variable(const.TYPES_BYTES),
                        Last(const.TYPES_INDEX),
                        First(const.TYPES_INDEX),
                    )
                    initialization = None
                if len(expression.elements) > 0:
                    aspects.append(
                        SizeAspect(Mul(Number(len(expression.elements)), Size(const.TYPES_BYTE)))
                    )

            result.global_declarations.append(
                ObjectDeclaration(
                    [ID(identifier)],
                    object_type,
                    initialization.substituted(self._substitution()).ada_expr()
                    if initialization
                    else None,
                    constant if initialization else False,
                    aspects,
                )
            )

        elif isinstance(type_, (rty.Integer, rty.Enumeration)):
            result.global_declarations.append(
                ObjectDeclaration(
                    [ID(identifier)],
                    self._ada_type(type_.identifier),
                    expression.ada_expr() if expression else None,
                )
            )

        elif isinstance(type_, (rty.Message, rty.Sequence)):
            if expression is not None:
                fail(
                    f"initialization for {type_} not yet supported",
                    Subsystem.GENERATOR,
                    location=expression.location,
                )

            type_identifier = self._ada_type(type_.identifier)

            result.global_declarations.append(
                self._declare_context(ID(identifier), type_identifier)
            )
            reset_message_contexts = IfStatement(
                [
                    (
                        Call(
                            ID(type_identifier) * "Has_Buffer",
                            [Variable(context_id(identifier))],
                        ),
                        self._free_context_buffer(
                            ID(identifier),
                            type_identifier,
                        ),
                    )
                ]
            )
            result.initialization_declarations.append(self._declare_buffer(ID(identifier)))
            result.initialization.extend(
                [
                    *([reset_message_contexts] if session_global else []),
                    self._allocate_buffer(identifier),
                    self._initialize_context(identifier, type_identifier),
                ]
            )
            result.finalization.extend(self._free_context_buffer(ID(identifier), type_identifier))

        else:
            fatal_fail(
                f"unexpected variable declaration for {type_}",
                Subsystem.GENERATOR,
                location=identifier.location,
            )

        assert isinstance(type_, (rty.Integer, rty.Enumeration, rty.Message, rty.Sequence))

        if session_global:
            self._session_context.referenced_types.add(type_.identifier)
        else:
            self._session_context.referenced_types_body.add(type_.identifier)

        return result

    def _declare_and_assign(
        self,
        variables: Sequence[Tuple[rid.ID, rty.Type, expr.Expr]],
        statements: Sequence[Statement],
        exception_handler: ExceptionHandler = None,
        constant: bool = False,
    ) -> Sequence[Statement]:
        if not variables:
            return statements

        k, type_, v = variables[0]
        initialized_in_declaration = isinstance(v, expr.Aggregate) and type_ == rty.OPAQUE
        evaluated_declaration = self._declare(
            k,
            type_,
            v if initialized_in_declaration else None,
            constant=constant,
        )

        return [
            Declare(
                [
                    *evaluated_declaration.global_declarations,
                    *evaluated_declaration.initialization_declarations,
                ],
                [
                    *[
                        *evaluated_declaration.initialization,
                        *(self._assign(k, type_, v) if not initialized_in_declaration else []),
                    ],
                    *self._declare_and_assign(
                        variables[1:], statements, exception_handler, constant
                    ),
                    *evaluated_declaration.finalization,
                ],
            ),
            *([self._realize_deferred_exception(exception_handler)] if exception_handler else []),
        ]

    def _assign(
        self,
        target: rid.ID,
        target_type: rty.Type,
        expression: expr.Expr,
        exception_handler: ExceptionHandler = None,
    ) -> Sequence[Statement]:
        if isinstance(expression, expr.Binding):
            return self._assign_to_binding(target, expression, exception_handler)

        if isinstance(expression, expr.Selected):
            return self._assign_to_selected(target, expression, exception_handler)

        if isinstance(expression, expr.MessageAggregate):
            return self._assign_to_message_aggregate(target, expression, exception_handler)

        if isinstance(expression, expr.Head):
            return self._assign_to_head(target, expression, exception_handler)

        if isinstance(expression, expr.Comprehension):
            assert isinstance(target_type, rty.Sequence)
            return self._assign_to_comprehension(target, target_type, expression, exception_handler)

        if isinstance(expression, expr.Call):
            return self._assign_to_call(target, expression, exception_handler)

        if isinstance(expression, expr.Conversion):
            return self._assign_to_conversion(target, expression, exception_handler)

        if isinstance(
            expression,
            (
                expr.Variable,
                expr.Number,
                expr.MathBinExpr,
                expr.MathAssExpr,
                expr.Relation,
                expr.Attribute,
                expr.Aggregate,
            ),
        ) and (
            isinstance(expression.type_, (rty.AnyInteger, rty.Enumeration))
            or expression.type_ == rty.OPAQUE
        ):
            return self._if_valid_fields(
                expression,
                [
                    Assignment(
                        ID(target),
                        expression.substituted(self._substitution()).ada_expr(),
                    )
                ],
                exception_handler,
            )

        if isinstance(expression, expr.Variable) and isinstance(
            expression.type_, (rty.Message, rty.Sequence)
        ):
            _unsupported_expression(expression, "in assignment")

        _unexpected_expression(expression, "in assignment")

    def _assign_to_binding(  # pylint: disable = too-many-branches
        self,
        target: rid.ID,
        binding: expr.Binding,
        exception_handler: ExceptionHandler = None,
    ) -> Sequence[Statement]:
        variables = []

        for k, v in binding.data.items():
            if isinstance(binding.expr, expr.MessageAggregate):
                assert isinstance(binding.expr.type_, rty.Message)
                for f, f_v in binding.expr.field_values.items():
                    if expr.Variable(k) == f_v:
                        type_ = binding.expr.type_.field_types[f]
                        break
                    if expr.Opaque(k) == f_v:
                        type_ = v.type_
                        break
                else:
                    fatal_fail(
                        f'"{k}" must be value of message aggregate',
                        Subsystem.GENERATOR,
                        location=binding.location,
                    )
            elif isinstance(binding.expr, expr.Call):
                for a, t in zip(binding.expr.args, binding.expr.argument_types):
                    if expr.Variable(k) == a:
                        type_ = t
                        break
                else:
                    fatal_fail(
                        f'"{k}" must be argument of call',
                        Subsystem.GENERATOR,
                        location=binding.location,
                    )
            elif isinstance(binding.expr, expr.Conversion):
                assert isinstance(binding.expr.argument, expr.Selected)
                assert isinstance(binding.expr.argument.prefix, expr.Variable)
                assert binding.expr.argument.prefix.identifier == k
                type_ = binding.expr.argument.prefix.type_
            elif isinstance(binding.expr, expr.Selected):
                assert isinstance(binding.expr.prefix, expr.Variable)
                assert binding.expr.prefix.identifier == k
                type_ = binding.expr.prefix.type_
            else:
                fail(
                    f'binding for expression "{type(binding.expr).__name__}" not yet supported',
                    Subsystem.GENERATOR,
                    location=binding.expr.location,
                )
            variables.append((k, type_, v))

        return self._declare_and_assign(
            variables,
            self._assign(target, type_, binding.expr),
            exception_handler,
            constant=True,
        )

    def _assign_to_selected(
        self, target: rid.ID, selected: expr.Selected, exception_handler: ExceptionHandler = None
    ) -> Sequence[Statement]:
        if not isinstance(selected.prefix, expr.Variable):
            fail(
                f'accessing field of expression "{type(selected.prefix).__name__}"'
                " not yet supported",
                Subsystem.GENERATOR,
                location=selected.prefix.location,
            )

        assert isinstance(selected.prefix.type_, rty.Message)

        message_type = ID(selected.prefix.type_.identifier)
        message_context = context_id(selected.prefix.identifier)
        selector = selected.selector

        if (
            isinstance(selected.type_, (rty.AnyInteger, rty.Enumeration))
            or selected.type_ == rty.OPAQUE
        ):
            return [
                IfStatement(
                    [
                        (
                            Call(
                                message_type * "Valid",
                                [
                                    Variable(message_context),
                                    Variable(message_type * f"F_{selector}"),
                                ],
                            ),
                            [
                                Assignment(
                                    Variable(ID(target)),
                                    Call(
                                        message_type * f"Get_{selector}",
                                        [Variable(message_context)],
                                    ),
                                )
                            ],
                        )
                    ],
                    # ISSUE: Componolit/RecordFlux#569
                    [
                        *self._debug_output(
                            f'Error: access to invalid field "{selector}" of'
                            f' "{message_context}"'
                        ),
                        *self._raise_exception(exception_handler),
                    ],
                )
            ]

        if isinstance(selected.type_, rty.Sequence):
            # ISSUE: Componolit/RecordFlux#577
            # The relevant buffer part has to be copied from the message context into a
            # sequence context. With the current implementation the sequence needs to
            # be parsed after copying. It must be ensured that the sequence is not
            # accidentally parsed beyond the original end of the sequence in the message
            # (i.e. misinterpreting trailing bytes in the new buffer as sequence elements).
            fail(
                "copying of sequence not yet supported",
                Subsystem.GENERATOR,
                location=target.location,
            )

        fatal_fail(
            f'unexpected type ({selected.type_}) for "{selected}" in assignment of "{target}"',
            Subsystem.GENERATOR,
            location=target.location,
        )

    def _assign_to_message_aggregate(
        self,
        target: rid.ID,
        message_aggregate: expr.MessageAggregate,
        exception_handler: ExceptionHandler = None,
    ) -> Sequence[Statement]:
        assert isinstance(message_aggregate.type_, rty.Message)

        self._session_context.used_types_body.add(const.TYPES_BIT_LENGTH)

        size = self._message_size(message_aggregate)
        size_variables = [
            v for v in size.variables() if isinstance(v.type_, (rty.Message, rty.Sequence))
        ]
        required_space = size.substituted(self._substitution()).ada_expr()
        target_type = ID(message_aggregate.type_.identifier)
        target_context = context_id(target)

        assign_to_message_aggregate = [
            self._if_sufficient_space(
                Call(
                    const.TYPES_BIT_LENGTH,
                    [
                        required_space,
                    ],
                ),
                target_context,
                [
                    CallStatement(
                        target_type * "Reset",
                        [
                            Variable(target_context),
                            Call(
                                const.TYPES_TO_FIRST_BIT_INDEX,
                                [
                                    Variable(f"{target_context}.Buffer_First"),
                                ],
                            ),
                            Add(
                                Call(
                                    const.TYPES_TO_FIRST_BIT_INDEX,
                                    [
                                        Variable(f"{target_context}.Buffer_First"),
                                    ],
                                ),
                                Call(
                                    const.TYPES_BIT_LENGTH,
                                    [
                                        required_space,
                                    ],
                                ),
                                -Number(1),
                            ),
                        ],
                    ),
                    *self._set_message_fields(target_type, target_context, message_aggregate),
                ],
                exception_handler,
            ),
        ]

        if size_variables:
            expressions = " or ".join(f'"{v}"' for v in size_variables)
            return [
                IfStatement(
                    [
                        (
                            AndThen(
                                *[
                                    e
                                    for v in size_variables
                                    if isinstance(v.type_, (rty.Message, rty.Sequence))
                                    for s in [
                                        expr.Size(v).substituted(self._substitution()).ada_expr()
                                    ]
                                    for e in [  # ISSUE: Componolit/RecordFlux#713
                                        LessEqual(s, Number(4096 * 8)),
                                        Equal(Mod(s, Size(const.TYPES_BYTE)), Number(0)),
                                    ]
                                ]
                            ),
                            assign_to_message_aggregate,
                        )
                    ],
                    # ISSUE: Componolit/RecordFlux#569
                    [
                        *self._debug_output(f"Error: unexpected size of {expressions}"),
                        *self._raise_exception(exception_handler),
                    ],
                )
            ]

        return assign_to_message_aggregate

    def _assign_to_head(
        self, target: rid.ID, head: expr.Head, exception_handler: ExceptionHandler = None
    ) -> Sequence[Statement]:
        assert isinstance(head.prefix.type_, rty.Sequence)

        if not isinstance(head.prefix.type_.element, (rty.Integer, rty.Enumeration, rty.Message)):
            fatal_fail(
                f"unexpected sequence element type ({head.prefix.type_.element})"
                f' for "{head}" in assignment of "{target}"',
                Subsystem.GENERATOR,
                location=head.location,
            )

        assert isinstance(head.type_, (rty.Integer, rty.Enumeration, rty.Message))

        if not isinstance(head.prefix, expr.Variable):
            _unsupported_expression(head.prefix, "in Head attribute")

        target_type = ID(head.type_.identifier)
        sequence_type = ID(head.prefix.type_.identifier)
        sequence_context = context_id(head.prefix.identifier)
        sequence_identifier = ID(rid.ID(f"{head.prefix}"))

        if isinstance(head.prefix.type_.element, (rty.Integer, rty.Enumeration)):
            return [
                IfStatement(
                    [
                        (
                            AndThen(
                                Call(sequence_type * "Valid", [Variable(sequence_context)]),
                                Call(
                                    sequence_type * "Has_Element",
                                    [Variable(sequence_context)],
                                ),
                                GreaterEqual(
                                    Call(
                                        sequence_type * "Size",
                                        [Variable(sequence_context)],
                                    ),
                                    Size(ID(head.prefix.type_.element.identifier)),
                                ),
                            ),
                            [
                                Assignment(
                                    Variable(ID(target)),
                                    Call(
                                        sequence_type * "Head",
                                        [Variable(sequence_context)],
                                    ),
                                )
                            ],
                        )
                    ],
                    # ISSUE: Componolit/RecordFlux#569
                    [
                        *self._debug_output(
                            f"Error: access to first element of invalid or empty sequence"
                            f' "{sequence_context}"'
                        ),
                        *self._raise_exception(exception_handler),
                    ],
                )
            ]

        assert isinstance(head.prefix.type_.element, rty.Message)

        self._session_context.used_types_body.add(const.TYPES_LENGTH)
        self._session_context.referenced_types_body.add(target_type)

        target_context = context_id(target)
        target_buffer = buffer_id(target)
        element_context = ID("RFLX_Head_Ctx")
        return [
            self._declare_sequence_copy(
                sequence_identifier,
                sequence_type,
                [
                    IfStatement(
                        [
                            (
                                Call(
                                    sequence_type * "Has_Element",
                                    [Variable(copy_id(sequence_context))],
                                ),
                                [
                                    Declare(
                                        [
                                            ObjectDeclaration(
                                                [element_context],
                                                target_type * "Context",
                                            ),
                                        ],
                                        [
                                            CallStatement(
                                                sequence_type * "Switch",
                                                [
                                                    Variable(copy_id(sequence_context)),
                                                    Variable(element_context),
                                                ],
                                            ),
                                            CallStatement(
                                                target_type * "Verify_Message",
                                                [Variable(element_context)],
                                            ),
                                            self._if_structural_valid_message(
                                                target_type,
                                                element_context,
                                                [
                                                    *self._take_buffer(
                                                        ID(target),
                                                        target_type,
                                                    ),
                                                    self._copy_to_buffer(
                                                        target_type,
                                                        element_context,
                                                        target_buffer,
                                                    ),
                                                    CallStatement(
                                                        target_type * "Initialize",
                                                        [
                                                            Variable(target_context),
                                                            Variable(target_buffer),
                                                        ],
                                                    ),
                                                    CallStatement(
                                                        target_type * "Verify_Message",
                                                        [
                                                            Variable(target_context),
                                                        ],
                                                    ),
                                                ],
                                            ),
                                            *self._update_context(
                                                copy_id(sequence_context),
                                                element_context,
                                                sequence_type,
                                            ),
                                        ],
                                    ),
                                ],
                            )
                        ],
                        [
                            # ISSUE: Componolit/RecordFlux#569
                            self._raise_deferred_exception(),
                        ],
                    )
                ],
            ),
            *([self._realize_deferred_exception(exception_handler)] if exception_handler else []),
        ]

    def _assign_to_comprehension(
        self,
        target: rid.ID,
        target_type: rty.Sequence,
        comprehension: expr.Comprehension,
        exception_handler: ExceptionHandler = None,
    ) -> Sequence[Statement]:
        assert isinstance(comprehension.type_, (rty.Sequence, rty.Aggregate))
        if not isinstance(comprehension.type_.element, (rty.Integer, rty.Enumeration)):
            fail(
                f"creating sequence with element {comprehension.type_.element}"
                " using list comprehension not yet supported",
                Subsystem.GENERATOR,
                location=comprehension.location,
            )
        assert isinstance(comprehension.sequence.type_, rty.Sequence)

        self._session_context.used_types_body.add(const.TYPES_BIT_LENGTH)

        target_id = ID(target)
        sequence_type_id = ID(comprehension.sequence.type_.identifier)
        iterator_id = ID(comprehension.iterator)

        sequence_element_type = comprehension.sequence.type_.element

        if isinstance(sequence_element_type, rty.Message):
            iterator_type_id = ID(sequence_element_type.identifier)

            if isinstance(comprehension.sequence, expr.Variable):
                sequence_id = ID(rid.ID(f"{comprehension.sequence}"))
                return [
                    self._declare_sequence_copy(
                        sequence_id,
                        sequence_type_id,
                        [
                            self._comprehension(
                                copy_id(sequence_id),
                                sequence_type_id,
                                target_id,
                                target_type,
                                iterator_id,
                                iterator_type_id,
                                comprehension.selector,
                                comprehension.condition,
                            )
                        ],
                    ),
                    *(
                        [self._realize_deferred_exception(exception_handler)]
                        if exception_handler
                        else []
                    ),
                ]

            if isinstance(comprehension.sequence, expr.Selected):
                selected = comprehension.sequence

                if not isinstance(selected.prefix, expr.Variable):
                    _unsupported_expression(
                        selected.prefix, "as prefix of Selected in list comprehension"
                    )

                assert isinstance(selected.prefix.type_, rty.Message)

                message_id = ID(selected.prefix.name)
                message_type = ID(selected.prefix.type_.identifier)
                sequence_id = ID(
                    rid.ID(
                        f"RFLX_{selected.prefix}_{selected.selector}",
                        location=selected.location,
                    )
                )
                message_field = ID(selected.selector)

                return [
                    self._if_structural_valid_message(
                        message_type,
                        context_id(message_id),
                        [
                            self._declare_message_field_sequence_copy(
                                message_id,
                                message_type,
                                message_field,
                                sequence_id,
                                sequence_type_id,
                                [
                                    self._comprehension(
                                        sequence_id,
                                        sequence_type_id,
                                        target_id,
                                        target_type,
                                        iterator_id,
                                        iterator_type_id,
                                        comprehension.selector,
                                        comprehension.condition,
                                    )
                                ],
                            ),
                        ],
                    ),
                    *(
                        [self._realize_deferred_exception(exception_handler)]
                        if exception_handler
                        else []
                    ),
                ]

            _unsupported_expression(comprehension.sequence, "as sequence in list comprehension")

        fail(
            f"iterating over sequence of {sequence_element_type}"
            " in list comprehension not yet supported",
            Subsystem.GENERATOR,
            location=comprehension.sequence.location,
        )

    def _assign_to_call(
        self,
        target: rid.ID,
        call_expr: expr.Call,
        exception_handler: ExceptionHandler = None,
    ) -> Sequence[Statement]:
        pre_call = []
        post_call = []
        local_declarations = []

        if isinstance(call_expr.type_, rty.Message):
            type_identifier = self._ada_type(call_expr.type_.identifier)
            local_declarations.append(
                ObjectDeclaration(
                    [ID(target)],
                    ID(type_identifier) * "Structure",
                )
            )
            post_call.append(
                CallStatement(
                    ID(type_identifier) * "To_Context",
                    [
                        Variable(ID(target)),
                        Variable(context_id(target)),
                    ],
                ),
            )

        for a in call_expr.args:
            if not isinstance(a, (expr.Number, expr.Variable, expr.Selected, expr.String)) and not (
                isinstance(a, expr.Opaque)
                and isinstance(a.prefix, expr.Variable)
                and a.type_ == rty.OPAQUE
            ):
                _unsupported_expression(a, "as function argument")

            if isinstance(a, expr.Variable) and isinstance(a.type_, rty.Message):
                type_identifier = self._ada_type(a.type_.identifier)
                local_declarations.append(
                    ObjectDeclaration(
                        [ID(a.identifier)],
                        ID(type_identifier) * "Structure",
                    )
                )
                pre_call.append(
                    CallStatement(
                        ID(type_identifier) * "To_Structure",
                        [
                            Variable(context_id(a.identifier)),
                            Variable(ID(a.identifier)),
                        ],
                    ),
                )

        call = [
            CallStatement(
                ID(call_expr.identifier),
                [
                    Variable(ID(target)),
                    *[a.substituted(self._substitution()).ada_expr() for a in call_expr.args],
                ],
            )
        ]

        if local_declarations:
            return self._if_valid_fields(
                expr.Or(*call_expr.args),
                [
                    Declare(
                        local_declarations,
                        [
                            *pre_call,
                            *call,
                            *post_call,
                        ],
                    )
                ],
                exception_handler,
            )

        return call

    @staticmethod
    def contains_function_name(
        refinement_package: rid.ID, pdu: rid.ID, sdu: rid.ID, field: rid.ID
    ) -> str:
        sdu_name = sdu.name if sdu.parent == refinement_package else sdu
        pdu_name = pdu.name if pdu.parent == refinement_package else pdu
        return f"{sdu_name.flat}_In_{pdu_name.flat}_{field}"

    def _assign_to_conversion(
        self,
        target: rid.ID,
        conversion: expr.Conversion,
        exception_handler: ExceptionHandler = None,
    ) -> Sequence[Statement]:
        assert isinstance(conversion.type_, rty.Message)
        assert isinstance(conversion.argument, expr.Selected)
        assert conversion.argument.type_ == rty.OPAQUE
        assert isinstance(conversion.argument.prefix, expr.Variable)
        assert isinstance(conversion.argument.prefix.type_, rty.Message)

        pdu = conversion.argument.prefix.type_
        sdu = conversion.type_
        field = conversion.argument.selector
        refinements = [
            r for r in pdu.refinements if r.sdu.identifier == sdu.identifier and r.field == field
        ]
        if not refinements:
            fatal_fail(
                f'no refinement for field "{field}" of message "{pdu.identifier}"'
                f' leads to "{sdu.identifier}"',
                Subsystem.GENERATOR,
                location=conversion.location,
            )
        assert len(refinements) == 1
        refinement = refinements[0]
        contains_package = refinement.package * "Contains"

        self._session_context.referenced_packages_body.add(contains_package)

        return [
            IfStatement(
                [
                    (
                        Call(
                            ID(contains_package)
                            * self.contains_function_name(
                                refinement.package, pdu.identifier, sdu.identifier, field
                            ),
                            [Variable(context_id(conversion.argument.prefix.identifier))],
                        ),
                        [
                            CallStatement(
                                ID(contains_package) * f"Copy_{field}",
                                [
                                    Variable(context_id(conversion.argument.prefix.identifier)),
                                    Variable(context_id(target)),
                                ],
                            ),
                            CallStatement(
                                ID(sdu.identifier) * "Verify_Message",
                                [
                                    Variable(context_id(target)),
                                ],
                            ),
                        ],
                    )
                ],
                # ISSUE: Componolit/RecordFlux#569
                [
                    *self._debug_output(f'Error: invalid conversion "{conversion}"'),
                    *self._raise_exception(exception_handler),
                ],
            )
        ]

    def _append(
        self,
        append: stmt.Append,
        exception_handler: ExceptionHandler,
    ) -> Sequence[Statement]:
        assert isinstance(append.type_, rty.Sequence)

        self._session_context.used_types_body.add(const.TYPES_BIT_LENGTH)

        def check(sequence_type: ID, required_space: Expr) -> Statement:
            return IfStatement(
                [
                    (
                        Or(
                            Not(
                                Call(
                                    sequence_type * "Has_Element",
                                    [Variable(sequence_context)],
                                )
                            ),
                            Less(
                                Call(
                                    sequence_type * "Available_Space",
                                    [Variable(sequence_context)],
                                ),
                                required_space,
                            ),
                        ),
                        [
                            *self._debug_output(
                                "Error: insufficient space for appending to sequence"
                                f' "{sequence_context}"'
                            ),
                            *exception_handler.execute(),
                        ],
                    )
                ]
            )

        if isinstance(append.type_.element, (rty.Integer, rty.Enumeration)):
            assert isinstance(append.parameter, expr.Variable)

            sequence_type = ID(append.type_.identifier)
            sequence_context = context_id(append.identifier)
            element_type = ID(append.type_.element.identifier)

            return [
                check(sequence_type, Size(element_type)),
                CallStatement(
                    sequence_type * "Append_Element",
                    [Variable(sequence_context), append.parameter.ada_expr()],
                ),
            ]

        if isinstance(append.type_.element, rty.Message):
            sequence_type = ID(append.type_.identifier)
            sequence_context = context_id(append.identifier)
            element_type = ID(append.type_.element.identifier)
            element_context = context_id("RFLX_Element_" + append.identifier)

            self._session_context.referenced_types_body.add(element_type)

            if isinstance(append.parameter, expr.MessageAggregate):
                required_space = (
                    self._message_size(append.parameter)
                    .substituted(self._substitution())
                    .ada_expr()
                )
            else:
                _unsupported_expression(append.parameter, "in Append statement")

            return [
                check(sequence_type, required_space),
                Declare(
                    [ObjectDeclaration([element_context], element_type * "Context")],
                    [
                        CallStatement(
                            sequence_type * "Switch",
                            [Variable(sequence_context), Variable(element_context)],
                        ),
                        *(
                            self._set_message_fields(
                                element_type,
                                element_context,
                                append.parameters[0],
                            )
                            if isinstance(append.parameters[0], expr.MessageAggregate)
                            else []
                        ),
                        *self._update_context(sequence_context, element_context, sequence_type),
                    ],
                ),
            ]

        fatal_fail(
            f"unexpected element type {append.type_.element} in Append statement",
            Subsystem.GENERATOR,
            location=append.parameter.location,
        )

    @staticmethod
    def _read(read: stmt.Read) -> Sequence[Statement]:
        assert isinstance(read.parameter.type_, rty.Message)

        if not isinstance(read.parameter, expr.Variable):
            _unsupported_expression(read.parameter, "in Read statement")

        target_type = ID(read.parameter.type_.identifier)
        target_context = context_id(read.parameter.identifier)
        identifier = f"{target_type.flat}_Write"
        return [
            Declare(
                [
                    GenericProcedureInstantiation(
                        identifier,
                        ProcedureSpecification(target_type * "Write"),
                        [f"{read.identifier}_Read"],
                    )
                ],
                [
                    CallStatement(identifier, [Variable(target_context)]),
                ],
            ),
            CallStatement(target_type * "Verify_Message", [Variable(target_context)]),
        ]

    def _write(
        self,
        write: stmt.Write,
        exception_handler: ExceptionHandler,
    ) -> Sequence[Statement]:
        assert isinstance(write.parameter.type_, rty.Message)

        if not isinstance(write.parameter, (expr.Variable, expr.MessageAggregate, expr.Binding)):
            _unsupported_expression(write.parameter, "in Write statement")

        if isinstance(write.parameter, expr.Variable):
            target_type = ID(write.parameter.type_.identifier)
            target_context = context_id(write.parameter.identifier)
        else:
            target_type = ID(write.parameter.type_.identifier)
            target_context = context_id(ID("RFLX_Message"))

        identifier = f"{target_type.flat}_Read"

        def write_statements(exception_handler: ExceptionHandler = None) -> Sequence[Statement]:
            return [
                self._if_structural_valid_message(
                    target_type,
                    target_context,
                    [
                        Declare(
                            [
                                GenericProcedureInstantiation(
                                    identifier,
                                    ProcedureSpecification(target_type * "Read"),
                                    [f"{write.identifier}_Write"],
                                )
                            ],
                            [
                                CallStatement(identifier, [Variable(target_context)]),
                            ],
                        )
                    ],
                    exception_handler,
                )
            ]

        if isinstance(write.parameter, expr.Variable):
            return write_statements(exception_handler)

        return self._declare_and_assign(
            [(ID("RFLX_Message"), write.parameter.type_, write.parameter)],
            write_statements(),
            exception_handler,
        )

    @staticmethod
    def _reset(
        reset: stmt.Reset,
    ) -> Sequence[Statement]:
        assert isinstance(reset.type_, rty.Message)

        target_type = ID(reset.type_.identifier)
        target_context = context_id(reset.identifier)
        return [
            CallStatement(target_type * "Reset", [Variable(target_context)]),
        ]

    def _message_size(self, message_aggregate: expr.MessageAggregate) -> expr.Expr:
        message = self._model_type(message_aggregate.identifier)
        assert isinstance(message, model.Message)
        return message.size({model.Field(f): v for f, v in message_aggregate.field_values.items()})

    def _substitution(self) -> Callable[[expr.Expr], expr.Expr]:
        # pylint: disable = too-many-statements

        def func(expression: expr.Expr) -> expr.Expr:
            # pylint: disable = too-many-branches, too-many-return-statements
            if isinstance(expression, (expr.Relation, expr.MathBinExpr)):
                for e in [expression.left, expression.right]:
                    if isinstance(e.type_, (rty.Integer, rty.Enumeration)):
                        self._session_context.used_types_body.add(e.type_.identifier)

            if isinstance(expression, expr.MathAssExpr):
                for e in expression.terms:
                    if isinstance(e.type_, (rty.Integer, rty.Enumeration)):
                        self._session_context.used_types_body.add(e.type_.identifier)

            if isinstance(expression, expr.And):
                return expr.AndThen(*expression.terms)

            if isinstance(expression, expr.Or):
                return expr.OrElse(*expression.terms)

            if isinstance(expression, expr.Selected):
                if isinstance(expression.prefix, expr.Variable):
                    assert isinstance(expression.prefix.type_, rty.Message)
                    return expr.Call(
                        ID(expression.prefix.type_.identifier) * ID(f"Get_{expression.selector}"),
                        [expr.Variable(context_id(expression.prefix.identifier))],
                    )

                assert False

            if isinstance(expression, expr.Valid):
                if isinstance(expression.prefix, expr.Variable):
                    if isinstance(expression.prefix.type_, rty.Message):
                        return expr.Call(
                            ID(expression.prefix.type_.identifier) * ID("Structural_Valid_Message"),
                            [expr.Variable(context_id(expression.prefix.identifier))],
                        )

                    if isinstance(expression.prefix.type_, rty.Sequence):
                        return expr.Call(
                            ID(expression.prefix.type_.identifier) * ID("Valid"),
                            [expr.Variable(context_id(expression.prefix.identifier))],
                        )

                    assert False

                if isinstance(expression.prefix, expr.Selected):
                    assert isinstance(expression.prefix.prefix, expr.Variable)
                    assert isinstance(expression.prefix.prefix.type_, rty.Message)
                    type_name = ID(expression.prefix.prefix.type_.identifier)
                    return expr.Call(
                        type_name
                        * (
                            ID("Valid")
                            if isinstance(expression.prefix.type_, (rty.Integer, rty.Enumeration))
                            else ID("Structural_Valid")
                        ),
                        [
                            expr.Variable(context_id(expression.prefix.prefix.identifier)),
                            expr.Variable(type_name * f"F_{expression.prefix.selector}"),
                        ],
                    )

                assert False

            if isinstance(expression, expr.Present):
                if isinstance(expression.prefix, expr.Selected):
                    assert isinstance(expression.prefix.prefix, expr.Variable)
                    assert isinstance(expression.prefix.prefix.type_, rty.Message)
                    type_name = ID(expression.prefix.prefix.type_.identifier)
                    return expr.Call(
                        type_name * ID("Present"),
                        [
                            expr.Variable(context_id(expression.prefix.prefix.identifier)),
                            expr.Variable(type_name * f"F_{expression.prefix.selector}"),
                        ],
                    )

                assert False

            if isinstance(expression, expr.Aggregate):
                assert len(expression.elements) > 0
                if len(expression.elements) == 1:
                    return expr.NamedAggregate(
                        (
                            str(expr.First(const.TYPES_INDEX)),
                            expr.Val(const.TYPES_BYTE, expression.elements[0]),
                        )
                    )
                return expr.Aggregate(*[expr.Val(const.TYPES_BYTE, e) for e in expression.elements])

            if (
                isinstance(expression, expr.Equal)
                and isinstance(expression.left, expr.Variable)
                and isinstance(expression.right, expr.Variable)
            ):
                if expression.left == expr.Variable("True"):
                    return expression.right
                if expression.right == expr.Variable("True"):
                    return expression.left
                if expression.left == expr.Variable("False"):
                    return expr.Not(expression.right)
                if expression.right == expr.Variable("False"):
                    return expr.Not(expression.left)

            if (
                isinstance(expression, expr.NotEqual)
                and isinstance(expression.left, expr.Variable)
                and isinstance(expression.right, expr.Variable)
            ):
                if expression.left == expr.Variable("True"):
                    return expr.Not(expression.right)
                if expression.right == expr.Variable("True"):
                    return expr.Not(expression.left)
                if expression.left == expr.Variable("False"):
                    return expression.right
                if expression.right == expr.Variable("False"):
                    return expression.left

            if isinstance(expression, (expr.Equal, expr.NotEqual)) and isinstance(
                expression.left.type_, rty.Enumeration
            ):
                relation = expression.__class__
                if expression.left.type_.always_valid:
                    selected = None
                    literal = None

                    if isinstance(expression.left, expr.Selected):
                        selected = expression.left.substituted(self._substitution())
                        literal = expression.right
                    elif isinstance(expression.right, expr.Selected):
                        selected = expression.right.substituted(self._substitution())
                        literal = expression.left

                    if selected and literal:
                        assert isinstance(literal.type_, rty.Enumeration)
                        self._session_context.used_types_body.add(
                            literal.type_.identifier + "_Enum"
                        )
                        return expr.AndThen(
                            expr.Selected(selected, "Known"),
                            relation(expr.Selected(selected, "Enum"), literal),
                        )

            if isinstance(expression, expr.Size):
                if isinstance(expression.prefix, expr.Variable):
                    if (
                        isinstance(expression.prefix.type_, rty.AnyInteger)
                        or (
                            isinstance(expression.prefix.type_, (rty.Sequence, rty.Aggregate))
                            and isinstance(expression.prefix.type_.element, rty.AnyInteger)
                        )
                        or (
                            isinstance(expression.prefix.type_, (rty.Integer, rty.Enumeration))
                            and expression.prefix.identifier == expression.prefix.type_.identifier
                        )
                    ):
                        return expression

                    if isinstance(expression.prefix.type_, (rty.Message, rty.Sequence)):
                        type_ = ID(expression.prefix.type_.identifier)
                        context = context_id(expression.prefix.identifier)
                        return expr.Call(type_ * "Size", [expr.Variable(context)])

                    _unexpected_expression(expression.prefix, "in Size attribute")

                if isinstance(expression.prefix, expr.Selected):
                    assert isinstance(expression.prefix.prefix, expr.Variable)
                    assert isinstance(expression.prefix.prefix.type_, rty.Message)
                    type_ = ID(expression.prefix.prefix.type_.identifier)
                    context = context_id(expression.prefix.prefix.identifier)
                    return expr.Call(
                        type_ * "Field_Size",
                        [
                            expr.Variable(context),
                            expr.Variable(type_ * "F_" + expression.prefix.selector),
                        ],
                    )

                _unsupported_expression(expression.prefix, "in Size attribute")

            if isinstance(expression, expr.HasData):
                assert isinstance(expression.prefix, expr.Variable)
                assert isinstance(expression.prefix.type_, rty.Channel)
                self._session_context.referenced_has_data.add(expression.prefix.identifier)
                return expr.Call(expression.prefix.identifier + "_Has_Data")

            if isinstance(expression, expr.Opaque):
                if isinstance(expression.prefix, expr.Variable):
                    assert expression.type_ == rty.OPAQUE
                    assert isinstance(expression.prefix.type_, rty.Message)
                    message_type = ID(expression.prefix.type_.identifier)
                    message_context = context_id(expression.prefix.identifier)
                    return expr.Call(
                        message_type * "Message_Data", [expr.Variable(message_context)]
                    )

                _unsupported_expression(expression.prefix, "in Opaque attribute")

            return expression

        return func

    def _if_valid_sequence(
        self,
        sequence_type: ID,
        sequence_context: ID,
        statements: Sequence[Statement],
        exception_handler: ExceptionHandler = None,
    ) -> IfStatement:
        """
        Ensure that sequence is valid.

        A deferred exception is raised in case of an invalid sequence.
        """
        # ISSUE: Componlit/RecordFlux#569
        return IfStatement(
            [
                (
                    Call(sequence_type * "Valid", [Variable(sequence_context)]),
                    statements,
                )
            ],
            # ISSUE: Componolit/RecordFlux#569
            [
                *self._debug_output(f'Error: invalid sequence "{sequence_context}"'),
                *self._raise_exception(exception_handler),
            ],
        )

    def _if_structural_valid_message(
        self,
        message_type: ID,
        message_context: ID,
        statements: Sequence[Statement],
        exception_handler: ExceptionHandler = None,
    ) -> IfStatement:
        """
        Ensure that message is structurally valid.

        An exception is raised in case of an invalid message. A deferred exceptions is assumed
        if state and finalization are not provided.
        """
        # ISSUE: Componlit/RecordFlux#569
        return IfStatement(
            [
                (
                    Call(
                        message_type * "Structural_Valid_Message",
                        [Variable(message_context)],
                    ),
                    statements,
                )
            ],
            # ISSUE: Componolit/RecordFlux#569
            [
                *self._debug_output(f'Error: invalid message "{message_context}"'),
                *self._raise_exception(exception_handler),
            ],
        )

    def _if_structural_valid_message_field(
        self,
        message_type: ID,
        message_context: ID,
        message_field: ID,
        statements: Sequence[Statement],
        exception_handler: ExceptionHandler = None,
    ) -> IfStatement:
        """
        Ensure that message field is structurally valid.

        A deferred exception is raised in case of an invalid message field.
        """
        # ISSUE: Componlit/RecordFlux#569
        return IfStatement(
            [
                (
                    Call(
                        message_type * "Structural_Valid",
                        [
                            Variable(message_context),
                            Variable(message_type * model.Field(message_field).affixed_name),
                        ],
                    ),
                    statements,
                )
            ],
            # ISSUE: Componolit/RecordFlux#569
            [
                *self._debug_output(
                    f'Error: invalid message field "{message_type * message_field}"'
                ),
                *self._raise_exception(exception_handler),
            ],
        )

    def _if_valid_fields(
        self,
        expression: expr.Expr,
        statements: Sequence[Statement],
        exception_handler: ExceptionHandler = None,
    ) -> Sequence[Statement]:
        """
        Ensure that all referenced fields in the expression are valid.

        An exception is raised in case of an invalid field. A deferred exception is raised
        if state and finalization are not provided.
        """
        # ISSUE: Componlit/RecordFlux#569

        selected = expression.findall(lambda x: isinstance(x, expr.Selected))

        if selected:
            expressions = " or ".join(f'"{s}"' for s in selected)
            return [
                IfStatement(
                    [
                        (
                            expr.AndThen(*[expr.Valid(e) for e in selected])
                            .substituted(self._substitution())
                            .ada_expr(),
                            statements,
                        )
                    ],
                    # ISSUE: Componolit/RecordFlux#569
                    [
                        *self._debug_output(
                            f"Error: reference to invalid message field in {expressions}"
                        ),
                        *self._raise_exception(exception_handler),
                    ],
                )
            ]

        return statements

    def _if_sufficient_space(
        self,
        required_space: Expr,
        target_context: ID,
        statements: Sequence[Statement],
        exception_handler: ExceptionHandler = None,
    ) -> IfStatement:
        return IfStatement(
            [
                (
                    GreaterEqual(
                        Add(
                            Variable(f"{target_context}.Last"),
                            -Variable(f"{target_context}.First"),
                            Number(1),
                        ),
                        required_space,
                    ),
                    statements,
                )
            ],
            [
                *self._debug_output(f'Error: insufficient space in message "{target_context}"'),
                *self._raise_exception(exception_handler),
            ],
        )

    def _if_sufficient_space_in_sequence(
        self,
        required_space: Expr,
        sequence_type: ID,
        sequence_context: ID,
        statements: Sequence[Statement],
        exception_handler: ExceptionHandler = None,
    ) -> IfStatement:
        return IfStatement(
            [
                (
                    AndThen(
                        Call(
                            sequence_type * "Has_Element",
                            [Variable(sequence_context)],
                        ),
                        GreaterEqual(
                            Call(
                                sequence_type * "Available_Space",
                                [Variable(sequence_context)],
                            ),
                            required_space,
                        ),
                    ),
                    statements,
                )
            ],
            [
                *self._debug_output(f'Error: insufficient space in sequence "{sequence_context}"'),
                *self._raise_exception(exception_handler),
            ],
        )

    def _raise_exception(self, exception_handler: ExceptionHandler = None) -> Sequence[Statement]:
        # ISSUE: Componolit/Workarounds#37
        return (
            exception_handler.execute() if exception_handler else [self._raise_deferred_exception()]
        )

    @staticmethod
    def _raise_deferred_exception() -> Assignment:
        # ISSUE: Componolit/Workarounds#37
        return Assignment("RFLX_Exception", TRUE)

    @staticmethod
    def _exit_on_deferred_exception() -> ExitStatement:
        # ISSUE: Componolit/Workarounds#37
        return ExitStatement(Variable("RFLX_Exception"))

    def _realize_deferred_exception(self, exception_handler: ExceptionHandler) -> IfStatement:
        """
        Execute the exception handler if a deferred exception was raised.

        This is needed because SPARK does not yet support return statements in the scope of a local
        owning declaration.
        """
        # ISSUE: Componolit/Workarounds#37
        self._session_context.state_exception.add(exception_handler.state.identifier)
        return IfStatement(
            [
                (
                    Variable("RFLX_Exception"),
                    exception_handler.execute(),
                )
            ]
        )

    def _set_message_fields(
        self,
        target_type: ID,
        target_context: ID,
        message_aggregate: expr.MessageAggregate,
    ) -> Sequence[Statement]:
        assert isinstance(message_aggregate.type_, rty.Message)

        result: List[Statement] = []

        for f, v in message_aggregate.field_values.items():
            if isinstance(v, (expr.Number, expr.Aggregate)) or (
                isinstance(v, (expr.Variable, expr.MathBinExpr, expr.MathAssExpr))
                and isinstance(v.type_, (rty.AnyInteger, rty.Enumeration, rty.Aggregate))
            ):
                field_type = message_aggregate.type_.field_types[f]
                if isinstance(v, expr.Aggregate) and len(v.elements) == 0:
                    result.append(
                        CallStatement(target_type * f"Set_{f}_Empty", [Variable(target_context)])
                    )
                else:
                    value = v.substituted(self._substitution())
                    if not field_type.is_compatible_strong(v.type_):
                        assert isinstance(field_type, rty.Integer)
                        value = expr.Conversion(field_type.identifier, value)
                    result.append(
                        CallStatement(
                            target_type * f"Set_{f}",
                            [
                                Variable(target_context),
                                value.ada_expr(),
                            ],
                        )
                    )
            elif isinstance(v, expr.Variable) and isinstance(v.type_, rty.Sequence):
                sequence_context = context_id(v.identifier)
                result.extend(
                    [
                        CallStatement(
                            target_type * f"Set_{f}",
                            [Variable(target_context), Variable(sequence_context)],
                        ),
                    ]
                )
            elif isinstance(v, expr.Variable) and isinstance(v.type_, rty.Message):
                _unsupported_expression(v, "in message aggregate")
            elif (
                isinstance(v, expr.Selected)
                and isinstance(v.prefix, expr.Variable)
                and isinstance(v.prefix.type_, rty.Message)
            ):
                message_type = ID(v.prefix.type_.identifier)
                message_context = context_id(v.prefix.identifier)
                target_field_type = message_aggregate.type_.field_types[f]
                assert isinstance(target_field_type, (rty.Integer, rty.Enumeration, rty.Sequence))
                get_field_value = Call(
                    message_type * f"Get_{v.selector}", [Variable(message_context)]
                )
                if not target_field_type.is_compatible_strong(v.type_):
                    get_field_value = Call(ID(target_field_type.identifier), [get_field_value])
                    self._session_context.referenced_types_body.add(
                        ID(target_field_type.identifier)
                    )
                result.extend(
                    [
                        CallStatement(
                            target_type * f"Set_{f}",
                            [
                                Variable(target_context),
                                get_field_value,
                            ],
                        ),
                    ]
                )
            elif isinstance(v, expr.Opaque) and isinstance(v.prefix, expr.Variable):
                assert v.type_ == rty.OPAQUE
                assert isinstance(v.prefix.type_, rty.Message)
                message_type = ID(v.prefix.type_.identifier)
                message_context = context_id(v.prefix.identifier)
                result.extend(
                    [
                        CallStatement(
                            target_type * f"Set_{f}",
                            [
                                Variable(target_context),
                                Call(message_type * "Message_Data", [Variable(message_context)]),
                            ],
                        ),
                    ]
                )
            else:
                _unsupported_expression(v, "in message aggregate")
        return result

    def _declare_context_buffer(self, identifier: ID, type_: ID) -> List[Declaration]:
        return [
            self._declare_context(identifier, type_),
            self._declare_buffer(identifier),
        ]

    def _declare_context(self, identifier: ID, type_: ID) -> Declaration:
        self._session_context.referenced_types_body.add(type_)
        return ObjectDeclaration([context_id(identifier)], type_ * "Context")

    @staticmethod
    def _declare_buffer(identifier: ID) -> Declaration:
        return ObjectDeclaration([buffer_id(identifier)], const.TYPES_BYTES_PTR)

    def _declare_sequence_copy(
        self,
        sequence_identifier: ID,
        sequence_type: ID,
        statements: Sequence[Statement],
        exception_handler: ExceptionHandler = None,
    ) -> IfStatement:
        """A deferred exception might be raised."""
        # ISSUE: Componolit/RecordFlux#577
        return self._if_valid_sequence(
            sequence_type,
            context_id(sequence_identifier),
            [
                Declare(
                    self._declare_context_buffer(copy_id(sequence_identifier), sequence_type),
                    [
                        self._allocate_buffer(copy_id(sequence_identifier)),
                        self._copy_to_buffer(
                            sequence_type,
                            context_id(sequence_identifier),
                            ID(f"{copy_id(buffer_id(sequence_identifier))}"),
                        ),
                        self._initialize_context(
                            copy_id(sequence_identifier),
                            sequence_type,
                            last=Call(
                                sequence_type * "Sequence_Last",
                                [Variable(context_id(sequence_identifier))],
                            ),
                        ),
                        *statements,
                        *self._free_context_buffer(copy_id(sequence_identifier), sequence_type),
                    ],
                )
            ],
            exception_handler,
        )

    def _declare_message_field_sequence_copy(  # pylint: disable = too-many-arguments
        self,
        message_identifier: ID,
        message_type: ID,
        message_field: ID,
        sequence_identifier: ID,
        sequence_type: ID,
        statements: Sequence[Statement],
    ) -> Declare:
        # ISSUE: Componolit/RecordFlux#577
        return Declare(
            self._declare_context_buffer(sequence_identifier, sequence_type),
            [
                self._allocate_buffer(sequence_identifier),
                self._copy_to_buffer(
                    message_type,
                    context_id(message_identifier),
                    ID(f"{buffer_id(sequence_identifier)}"),
                ),
                self._if_structural_valid_message_field(
                    message_type,
                    context_id(message_identifier),
                    message_field,
                    [
                        self._initialize_context(
                            sequence_identifier,
                            sequence_type,
                            first=Call(
                                message_type * "Field_First",
                                [
                                    Variable(context_id(message_identifier)),
                                    Variable(
                                        message_type * model.Field(message_field).affixed_name
                                    ),
                                ],
                            ),
                            last=Call(
                                message_type * "Field_Last",
                                [
                                    Variable(context_id(message_identifier)),
                                    Variable(
                                        message_type * model.Field(message_field).affixed_name
                                    ),
                                ],
                            ),
                        ),
                        *statements,
                        *self._take_buffer(sequence_identifier, sequence_type),
                    ],
                ),
                self._free_buffer(sequence_identifier),
            ],
        )

    def _comprehension(  # pylint: disable = too-many-arguments
        self,
        sequence_identifier: ID,
        sequence_type: ID,
        target_identifier: ID,
        target_type: rty.Sequence,
        iterator_identifier: ID,
        iterator_type: ID,
        selector: expr.Expr,
        condition: expr.Expr,
    ) -> While:
        """A deferred exception might be raised."""
        assert not isinstance(selector, expr.MessageAggregate)

        assert isinstance(target_type.element, (rty.Integer, rty.Enumeration, rty.Message))

        target_type_id = ID(target_type.identifier)

        return While(
            Call(
                sequence_type * "Has_Element",
                [Variable(context_id(sequence_identifier))],
            ),
            [
                PragmaStatement(
                    "Loop_Invariant",
                    [
                        Call(
                            sequence_type * "Has_Buffer",
                            [Variable(context_id(sequence_identifier))],
                        )
                    ],
                ),
                PragmaStatement(
                    "Loop_Invariant",
                    [
                        Call(
                            target_type_id * "Has_Buffer",
                            [Variable(context_id(target_identifier))],
                        )
                    ],
                ),
                PragmaStatement(
                    "Loop_Invariant",
                    [
                        Call(
                            target_type_id * "Valid",
                            [Variable(context_id(target_identifier))],
                        )
                    ],
                ),
                Declare(
                    [self._declare_context(iterator_identifier, iterator_type)],
                    [
                        CallStatement(
                            sequence_type * "Switch",
                            [
                                Variable(context_id(sequence_identifier)),
                                Variable(context_id(iterator_identifier)),
                            ],
                        ),
                        CallStatement(
                            iterator_type * "Verify_Message",
                            [Variable(context_id(iterator_identifier))],
                        ),
                        *self._if_valid_fields(
                            condition,
                            [
                                IfStatement(
                                    [
                                        (
                                            condition.substituted(self._substitution()).ada_expr(),
                                            [
                                                self._if_sufficient_space_in_sequence(
                                                    Size(
                                                        ID(
                                                            target_type.element.identifier + "_Enum"
                                                            if isinstance(
                                                                target_type.element, rty.Enumeration
                                                            )
                                                            and target_type.element.always_valid
                                                            else target_type.element.identifier
                                                        )
                                                    ),
                                                    target_type_id,
                                                    context_id(target_identifier),
                                                    [
                                                        CallStatement(
                                                            target_type_id * "Append_Element",
                                                            [
                                                                Variable(
                                                                    context_id(target_identifier)
                                                                ),
                                                                selector.substituted(
                                                                    self._substitution()
                                                                ).ada_expr(),
                                                            ],
                                                        )
                                                    ],
                                                ),
                                            ],
                                        )
                                    ]
                                )
                            ],
                        ),
                        *self._update_context(
                            context_id(sequence_identifier),
                            context_id(iterator_identifier),
                            sequence_type,
                        ),
                    ],
                ),
                self._exit_on_deferred_exception(),
            ],
        )

    def _free_context_buffer(self, identifier: ID, type_: ID) -> Sequence[Statement]:
        return [
            *self._take_buffer(identifier, type_),
            self._free_buffer(identifier),
        ]

    @staticmethod
    def _free_buffer(identifier: ID) -> Statement:
        return CallStatement(
            const.TYPES * "Free",
            [Variable(buffer_id(identifier))],
        )

    @staticmethod
    def _take_buffer(identifier: ID, type_: ID) -> Sequence[Statement]:
        context = context_id(identifier)
        return [
            # WORKAROUND: Componolit/Workarounds#32
            PragmaStatement(
                "Warnings",
                [
                    Variable("Off"),
                    String(f'unused assignment to "{context}"'),
                ],
            ),
            PragmaStatement(
                "Warnings",
                [
                    Variable("Off"),
                    String(f'"{context}" is set by "Take_Buffer" but not used after the call'),
                ],
            ),
            CallStatement(
                type_ * "Take_Buffer",
                [
                    Variable(context),
                    Variable(buffer_id(identifier)),
                ],
            ),
            PragmaStatement(
                "Warnings",
                [
                    Variable("On"),
                    String(f'"{context}" is set by "Take_Buffer" but not used after the call'),
                ],
            ),
            PragmaStatement(
                "Warnings",
                [
                    Variable("On"),
                    String(f'unused assignment to "{context}"'),
                ],
            ),
        ]

    @staticmethod
    def _update_context(
        sequence_context: ID, element_context: ID, sequence_type: ID
    ) -> Sequence[Statement]:
        return [
            # WORKAROUND: Componolit/Workarounds#32
            PragmaStatement(
                "Warnings",
                [
                    Variable("Off"),
                    String(f'unused assignment to "{element_context}"'),
                ],
            ),
            PragmaStatement(
                "Warnings",
                [
                    Variable("Off"),
                    String(f'"{element_context}" is set by "Update" but not used after the call'),
                ],
            ),
            CallStatement(
                sequence_type * "Update",
                [
                    Variable(sequence_context),
                    Variable(element_context),
                ],
            ),
            PragmaStatement(
                "Warnings",
                [
                    Variable("On"),
                    String(f'"{element_context}" is set by "Update" but not used after the call'),
                ],
            ),
            PragmaStatement(
                "Warnings",
                [
                    Variable("On"),
                    String(f'unused assignment to "{element_context}"'),
                ],
            ),
        ]

    @staticmethod
    def _allocate_buffer(identifier: rid.ID, initialization: Expr = None) -> Assignment:
        return Assignment(
            buffer_id(identifier),
            New(
                QualifiedExpr(
                    const.TYPES_BYTES,
                    initialization
                    if initialization
                    else NamedAggregate(
                        (  # ISSUE: Componolit/RecordFlux#713
                            ValueRange(
                                First(const.TYPES_INDEX),
                                Add(First(const.TYPES_INDEX), Number(4095)),
                            ),
                            First(const.TYPES_BYTE),
                        )
                    ),
                )
            ),
        )

    @staticmethod
    def _initialize_context(
        identifier: rid.ID, type_: ID, first: Expr = None, last: Expr = None
    ) -> CallStatement:
        return CallStatement(
            type_ * "Initialize",
            [
                Variable(context_id(identifier)),
                Variable(buffer_id(identifier)),
                *(
                    [
                        first
                        or Call(const.TYPES_TO_FIRST_BIT_INDEX, [First(buffer_id(identifier))]),
                        last or Call(const.TYPES_TO_FIRST_BIT_INDEX, [Last(buffer_id(identifier))]),
                    ]
                    if first or last
                    else []
                ),
            ],
        )

    def _copy_to_buffer(self, type_: ID, source_context: ID, target_buffer: ID) -> IfStatement:
        """A deferred exception might be raised."""
        self._session_context.used_types_body.add(const.TYPES_LENGTH)
        return IfStatement(
            [
                (
                    LessEqual(
                        Call(
                            type_ * "Byte_Size",
                            [Variable(source_context)],
                        ),
                        Length(target_buffer),
                    ),
                    [
                        CallStatement(
                            type_ * "Copy",
                            [
                                Variable(source_context),
                                Indexed(
                                    Variable(target_buffer * "all"),
                                    ValueRange(
                                        First(target_buffer),
                                        Add(
                                            First(target_buffer),
                                            Call(
                                                const.TYPES_INDEX,
                                                [
                                                    Add(
                                                        Call(
                                                            type_ * "Byte_Size",
                                                            [Variable(source_context)],
                                                        ),
                                                        Number(1),
                                                    )
                                                ],
                                            ),
                                            -Number(2),
                                        ),
                                    ),
                                ),
                            ],
                        ),
                    ],
                )
            ],
            [
                self._raise_deferred_exception(),
            ],
        )

    def _debug_output(self, string: str) -> List[CallStatement]:
        return [CallStatement("Ada.Text_IO.Put_Line", [String(string)])] if self._debug else []


def copy_id(identifier: ID) -> ID:
    return "RFLX_Copy_" + identifier


def context_id(identifier: rid.ID) -> ID:
    return ID(identifier + "_Ctx")


def buffer_id(identifier: rid.ID) -> ID:
    return ID(identifier + "_Buffer")


def _unexpected_expression(expression: expr.Expr, context: str) -> NoReturn:
    fatal_fail(
        f"unexpected {type(expression).__name__} with {expression.type_} {context}",
        Subsystem.GENERATOR,
        location=expression.location,
    )


def _unsupported_expression(expression: expr.Expr, context: str) -> NoReturn:
    fail(
        f"{type(expression).__name__} with {expression.type_} {context} not yet supported",
        Subsystem.GENERATOR,
        location=expression.location,
    )



# pylint: disable=too-many-lines

import itertools
from abc import abstractmethod
from collections import defaultdict
from copy import copy
from dataclasses import dataclass, field as dataclass_field
from typing import Dict, List, Mapping, Optional, Sequence, Set, Tuple, Union

import rflx.typing_ as rty
from rflx import expression as expr
from rflx.common import Base, indent, indent_next, unique, verbose_repr
from rflx.contract import ensure, invariant
from rflx.error import Location, RecordFluxError, Severity, Subsystem, fail, fatal_fail
from rflx.identifier import ID, StrID

from . import type_ as mty


class Field(Base):
    def __init__(self, identifier: StrID) -> None:
        self.identifier = ID(identifier)

    def __hash__(self) -> int:
        return hash(self.identifier)

    def __repr__(self) -> str:
        return f'Field("{self.identifier}")'

    def __lt__(self, other: "Field") -> int:
        return self.identifier < other.identifier

    @property
    def name(self) -> str:
        return str(self.identifier)

    @property
    def affixed_name(self) -> str:
        return f"F_{self.name}"


INITIAL = Field("Initial")
FINAL = Field("Final")


@dataclass(order=True)
class Link(Base):
    source: Field
    target: Field
    condition: expr.Expr = expr.TRUE
    size: expr.Expr = expr.UNDEFINED
    first: expr.Expr = expr.UNDEFINED
    location: Optional[Location] = dataclass_field(default=None, repr=False)

    def __str__(self) -> str:
        condition = indent_next(
            f"\nif {indent_next(str(self.condition), 3)}" if self.condition != expr.TRUE else "", 3
        )
        aspects = []
        if self.size != expr.UNDEFINED:
            aspects.append(f"Size => {self.size}")
        if self.first != expr.UNDEFINED:
            aspects.append(f"First => {self.first}")
        with_clause = indent_next("\nwith " + ", ".join(aspects) if aspects else "", 3)
        target_name = self.target.name if self.target != FINAL else "null"
        return f"then {target_name}{with_clause}{condition}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.source == other.source
                and self.target == other.target
                and self.condition == other.condition
                and self.size == other.size
                and self.first == other.first
            )
        return NotImplemented

    def __hash__(self) -> int:
        return 0


def valid_message_field_types(message: "AbstractMessage") -> bool:
    for t in message.types.values():
        if not isinstance(t, (mty.Scalar, mty.Composite, AbstractMessage)):
            return False
    return True


class MessageState(Base):
    fields: Optional[Tuple[Field, ...]] = ()
    definite_predecessors: Optional[Mapping[Field, Tuple[Field, ...]]] = None
    field_condition: Optional[Mapping[Field, expr.Expr]] = None
    checksums: Mapping[ID, Sequence[expr.Expr]] = {}


@invariant(lambda self: valid_message_field_types(self))
@invariant(lambda self: not self.types if not self.structure else True)
class AbstractMessage(mty.Type):
    # pylint: disable=too-many-arguments,too-many-public-methods,too-many-instance-attributes
    def __init__(
        self,
        identifier: StrID,
        structure: Sequence[Link],
        types: Mapping[Field, mty.Type],
        aspects: Mapping[ID, Mapping[ID, Sequence[expr.Expr]]] = None,
        location: Location = None,
        error: RecordFluxError = None,
        state: MessageState = None,
    ) -> None:
        assert not (state and aspects)

        super().__init__(identifier, location, error)

        self.structure = sorted(structure)
        self.__types = types
        self.__aspects = aspects or {}
        self.__has_unreachable = False
        self._state = state or MessageState()
        self.__paths_cache: Dict[Field, Set[Tuple[Link, ...]]] = {}

        assert len(self.identifier.parts) > 1, "type identifier must contain package"

        self._unqualified_enum_literals = {
            l
            for t in self.dependencies
            if isinstance(t, mty.Enumeration) and t.package == self.package
            for l in t.literals
        }
        self._qualified_enum_literals = mty.qualified_enum_literals(self.dependencies)
        self._type_literals = mty.qualified_type_literals(self.dependencies)

        if not state and (structure or types):
            try:
                self.__validate()
                self.__normalize()
                self._state.fields = self.__compute_topological_sorting()
                if self._state.fields:
                    self.__types = {f: self.__types[f] for f in self._state.fields}
                if ID("Checksum") in self.__aspects:
                    self._state.checksums = self.__aspects[ID("Checksum")]
            except RecordFluxError:
                pass

    def __hash__(self) -> int:
        return hash(self.identifier)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.identifier == other.identifier
                and self.structure == other.structure
                and self.types == other.types
                and self.aspects == other.aspects
            )
        return NotImplemented

    def __repr__(self) -> str:
        return verbose_repr(self, ["identifier", "structure", "types"])

    def __str__(self) -> str:
        if not self.structure or not self.types:
            return f"type {self.name} is null message"

        fields = ""
        field_list = [INITIAL, *self.fields]
        for i, f in enumerate(field_list):
            if f != INITIAL:
                fields += "\n" if fields else ""
                fields += f"{f.name} : {self.types[f].name}"
            outgoing = self.outgoing(f)
            if not (
                len(outgoing) == 1
                and outgoing[0].condition == expr.TRUE
                and outgoing[0].size == expr.UNDEFINED
                and outgoing[0].first == expr.UNDEFINED
                and (i >= len(field_list) - 1 or field_list[i + 1] == outgoing[0].target)
            ):
                if f == INITIAL:
                    fields += "null"
                fields += "\n" + indent("\n".join(str(o) for o in outgoing), 3)
            if fields:
                fields += ";"
        return f"type {self.name} is\n   message\n{indent(fields, 6)}\n   end message"

    @property
    def dependencies(self) -> List[mty.Type]:
        return [self, *unique(a for t in self.types.values() for a in t.dependencies)]

    @abstractmethod
    def copy(
        self,
        identifier: StrID = None,
        structure: Sequence[Link] = None,
        types: Mapping[Field, mty.Type] = None,
        aspects: Mapping[ID, Mapping[ID, Sequence[expr.Expr]]] = None,
        location: Location = None,
        error: RecordFluxError = None,
    ) -> "AbstractMessage":
        raise NotImplementedError

    @abstractmethod
    def proven(self, skip_proof: bool = False) -> "Message":
        raise NotImplementedError

    @property
    def fields(self) -> Tuple[Field, ...]:
        """Return fields topologically sorted."""
        return self._state.fields or ()

    @property
    def all_fields(self) -> Tuple[Field, ...]:
        return (INITIAL, *self.fields, FINAL)

    @property
    def types(self) -> Mapping[Field, mty.Type]:
        """Return fields and corresponding types topologically sorted."""
        return self.__types

    @property
    def aspects(self) -> Mapping[ID, Mapping[ID, Sequence[expr.Expr]]]:
        return self.__aspects

    @property
    def checksums(self) -> Mapping[ID, Sequence[expr.Expr]]:
        return self._state.checksums or {}

    def incoming(self, field: Field) -> List[Link]:
        return [l for l in self.structure if l.target == field]

    def outgoing(self, field: Field) -> List[Link]:
        return [l for l in self.structure if l.source == field]

    def predecessors(self, field: Field) -> Tuple[Field, ...]:
        if field == INITIAL:
            return ()
        if field == FINAL:
            return self.fields
        return self.fields[: self.fields.index(field)]

    def successors(self, field: Field) -> Tuple[Field, ...]:
        if field == INITIAL:
            return self.fields
        if field == FINAL:
            return ()
        return self.fields[self.fields.index(field) + 1 :]

    def direct_predecessors(self, field: Field) -> List[Field]:
        return list(dict.fromkeys([l.source for l in self.incoming(field)]))

    def direct_successors(self, field: Field) -> List[Field]:
        return list(dict.fromkeys([l.target for l in self.outgoing(field)]))

    def definite_predecessors(self, field: Field) -> Tuple[Field, ...]:
        """Return preceding fields which are part of all possible paths."""
        if self._state.definite_predecessors is None:
            self._state.definite_predecessors = {
                f: self.__compute_definite_predecessors(f) for f in self.all_fields
            }
        return self._state.definite_predecessors[field]

    def field_condition(self, field: Field) -> expr.Expr:
        if self._state.field_condition is None:
            self._state.field_condition = {
                f: self.__compute_field_condition(f).simplified() for f in self.all_fields
            }
        return self._state.field_condition[field]

    def field_size(self, field: Field) -> expr.Number:
        """Return field size if field size is fixed and fail otherwise."""
        if field == FINAL:
            return expr.Number(0)

        assert field in self.fields, f'field "{field.name}" not found'

        field_type = self.types[field]

        if isinstance(field_type, mty.Scalar):
            return field_type.size

        sizes = [
            l.size.substituted(mapping=to_mapping(self.type_constraints(expr.TRUE))).simplified()
            for l in self.incoming(field)
        ]
        size = sizes[0]
        if isinstance(size, expr.Number) and all(size == s for s in sizes):
            return size

        fail(
            f'unable to calculate size of field "{field.name}" of message "{self.identifier}"',
            Subsystem.MODEL,
            Severity.ERROR,
            field.identifier.location,
        )

    def paths(self, field: Field) -> Set[Tuple[Link, ...]]:
        if field == INITIAL:
            return set()
        if field in self.__paths_cache:
            return self.__paths_cache[field]

        result = set()
        for l in self.incoming(field):
            source = self.paths(l.source)
            for s in source:
                result.add(s + (l,))
            if not source:
                result.add((l,))

        self.__paths_cache[field] = result
        return result

    def prefixed(self, prefix: str) -> "AbstractMessage":
        fields = {f.identifier for f in self.fields}

        def prefixed_expression(expression: expr.Expr) -> expr.Expr:
            variables = {v.identifier for v in expression.variables()}
            literals = {l for l in variables - fields if len(l.parts) == 1}

            return expression.substituted(
                mapping={
                    **{
                        v: v.__class__(ID(prefix) + v.name)
                        for v in expression.variables()
                        if v.identifier in fields
                    },
                    **{
                        v: v.__class__(self.package * v.name)
                        for v in expression.variables()
                        if v.identifier in literals
                        and v.identifier not in mty.BUILTIN_LITERALS
                        and v.identifier != ID("Message")
                    },
                }
            ).simplified()

        structure = []

        for l in self.structure:
            source = Field(prefix + l.source.identifier) if l.source != INITIAL else INITIAL
            target = Field(prefix + l.target.identifier) if l.target != FINAL else FINAL
            condition = prefixed_expression(l.condition)
            size = prefixed_expression(l.size)
            first = prefixed_expression(l.first)
            structure.append(Link(source, target, condition, size, first, l.location))

        types = {Field(prefix + f.identifier): t for f, t in self.types.items()}

        return self.copy(structure=structure, types=types)

    def type_constraints(self, expression: expr.Expr) -> List[expr.Expr]:
        def get_constraints(aggregate: expr.Aggregate, field: expr.Variable) -> Sequence[expr.Expr]:
            comp = self.types[Field(field.name)]
            assert isinstance(comp, mty.Composite)
            result = expr.Equal(
                expr.Mul(aggregate.length, comp.element_size),
                expr.Size(field),
                location=expression.location,
            )
            if isinstance(comp, mty.Sequence) and isinstance(comp.element_type, mty.Scalar):
                return [
                    result,
                    *comp.element_type.constraints(name=comp.element_type.name, proof=True),
                ]
            return [result]

        scalar_types = [
            (f.name, t)
            for f, t in self.types.items()
            if isinstance(t, mty.Scalar)
            and ID(f.name) not in self._qualified_enum_literals
            and f.name not in ["Message", "Final"]
        ]

        aggregate_constraints: List[expr.Expr] = []
        for r in expression.findall(lambda x: isinstance(x, (expr.Equal, expr.NotEqual))):
            assert isinstance(r, (expr.Equal, expr.NotEqual))
            if isinstance(r.left, expr.Aggregate) and isinstance(r.right, expr.Variable):
                aggregate_constraints.extend(get_constraints(r.left, r.right))
            if isinstance(r.left, expr.Variable) and isinstance(r.right, expr.Aggregate):
                aggregate_constraints.extend(get_constraints(r.right, r.left))

        scalar_constraints = [
            c
            for n, t in scalar_types
            for c in t.constraints(name=n, proof=True, same_package=False)
        ]

        type_size_constraints = [
            expr.Equal(expr.Size(l), t.size)
            for l, t in self._type_literals.items()
            if isinstance(t, mty.Scalar)
        ]

        return [
            *aggregate_constraints,
            *scalar_constraints,
            *type_size_constraints,
        ]

    @classmethod
    def message_constraints(cls) -> List[expr.Expr]:
        return [
            expr.Equal(expr.Mod(expr.First("Message"), expr.Number(8)), expr.Number(1)),
            expr.Equal(expr.Mod(expr.Size("Message"), expr.Number(8)), expr.Number(0)),
        ]

    def __validate(self) -> None:
        # pylint: disable=too-many-branches, too-many-locals

        type_fields = self.__types.keys() | {INITIAL, FINAL}
        structure_fields = {l.source for l in self.structure} | {l.target for l in self.structure}

        for f in structure_fields - type_fields:
            self.error.append(
                f'missing type for field "{f.name}" in "{self.identifier}"',
                Subsystem.MODEL,
                Severity.ERROR,
                f.identifier.location,
            )

        for f in type_fields - structure_fields - {FINAL}:
            self.error.append(
                f'unused field "{f.name}" in "{self.identifier}"',
                Subsystem.MODEL,
                Severity.ERROR,
                f.identifier.location,
            )

        initial_links = self.outgoing(INITIAL)

        if len(initial_links) != 1:
            self.error.append(
                f'ambiguous first field in "{self.identifier}"',
                Subsystem.MODEL,
                Severity.ERROR,
                self.location,
            )
            self.error.extend(
                [
                    ("duplicate", Subsystem.MODEL, Severity.INFO, l.target.identifier.location)
                    for l in self.outgoing(INITIAL)
                    if l.target.identifier.location
                ]
            )

        if initial_links[0].first != expr.UNDEFINED:
            self.error.append(
                "illegal first aspect at initial link",
                Subsystem.MODEL,
                Severity.ERROR,
                initial_links[0].first.location,
            )

        name_conflicts = [
            (f, l)
            for f in type_fields
            for l in self._unqualified_enum_literals
            if f.identifier == l
        ]

        if name_conflicts:
            conflicting_field, conflicting_literal = name_conflicts.pop(0)
            self.error.append(
                f'name conflict for field "{conflicting_field.name}" in "{self.identifier}"',
                Subsystem.MODEL,
                Severity.ERROR,
                conflicting_field.identifier.location,
            )
            self.error.append(
                "conflicting enumeration literal",
                Subsystem.MODEL,
                Severity.INFO,
                conflicting_literal.location,
            )

        self.error.propagate()

        for f in structure_fields:
            for l in self.structure:
                if f in (INITIAL, l.target):
                    break
            else:
                self.__has_unreachable = True
                self.error.append(
                    f'unreachable field "{f.name}" in "{self.identifier}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    f.identifier.location,
                )

        duplicate_links = defaultdict(list)
        for link in self.structure:
            duplicate_links[(link.source, link.target, link.condition)].append(link)

        for links in duplicate_links.values():
            if len(links) > 1:
                self.error.append(
                    f'duplicate link from "{links[0].source.identifier}"'
                    f' to "{links[0].target.identifier}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    links[0].source.identifier.location,
                )
                self.error.extend(
                    [
                        (
                            "duplicate link",
                            Subsystem.MODEL,
                            Severity.INFO,
                            l.location,
                        )
                        for l in links
                    ]
                )

        for l in self.structure:
            exponentiations = itertools.chain.from_iterable(
                e.findall(lambda x: isinstance(x, expr.Pow)) for e in [l.condition, l.first, l.size]
            )
            for e in exponentiations:
                assert isinstance(e, expr.Pow)
                variables = e.right.findall(lambda x: isinstance(x, expr.Variable))
                if variables:
                    self.error.append(
                        f'unsupported expression in "{self.identifier}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        e.location,
                    )
                    for v in variables:
                        self.error.append(
                            f'variable "{v}" in exponent',
                            Subsystem.MODEL,
                            Severity.INFO,
                            v.location,
                        )

    def __normalize(self) -> None:
        """Qualify enumeration literals in conditions to prevent ambiguities."""

        def qualify_enum_literals(expression: expr.Expr) -> expr.Expr:
            if (
                isinstance(expression, expr.Variable)
                and expression.identifier in self._unqualified_enum_literals
            ):
                return expr.Variable(
                    self.package * expression.identifier,
                    negative=expression.negative,
                    type_=expression.type_,
                    location=expression.location,
                )
            return expression

        for link in self.structure:
            link.condition = link.condition.substituted(qualify_enum_literals)

    def __compute_topological_sorting(self) -> Optional[Tuple[Field, ...]]:
        """Return fields topologically sorted (Kahn's algorithm)."""
        result: Tuple[Field, ...] = ()
        fields = [INITIAL]
        visited = set()
        while fields:
            n = fields.pop(0)
            result += (n,)
            for e in self.outgoing(n):
                visited.add(e)
                if set(self.incoming(e.target)) <= visited:
                    fields.append(e.target)
        if not self.__has_unreachable and set(self.structure) - visited:
            self.error.append(
                f'structure of "{self.identifier}" contains cycle',
                Subsystem.MODEL,
                Severity.ERROR,
                self.location,
            )
            # ISSUE: Componolit/RecordFlux#256
            return None
        return tuple(f for f in result if f not in [INITIAL, FINAL])

    def __compute_definite_predecessors(self, final: Field) -> Tuple[Field, ...]:
        return tuple(
            f
            for f in self.fields
            if all(any(f == pf.source for pf in p) for p in self.paths(final))
        )

    def __compute_field_condition(self, final: Field) -> expr.Expr:
        if final == INITIAL:
            return expr.TRUE
        return expr.Or(
            *[
                expr.And(self.__compute_field_condition(l.source), l.condition)
                for l in self.incoming(final)
            ],
            location=final.identifier.location,
        )


class Message(AbstractMessage):
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        identifier: StrID,
        structure: Sequence[Link],
        types: Mapping[Field, mty.Type],
        aspects: Mapping[ID, Mapping[ID, Sequence[expr.Expr]]] = None,
        location: Location = None,
        error: RecordFluxError = None,
        state: MessageState = None,
        skip_proof: bool = False,
    ) -> None:
        super().__init__(identifier, structure, types, aspects, location, error, state)

        self._refinements: List["Refinement"] = []

        if not self.error.check() and not skip_proof:
            self.verify()

        self.error.propagate()

    def verify(self) -> None:
        if self.structure or self.types:
            self.__verify_expression_types()
            self.__verify_expressions()
            self.__verify_checksums()

            self.error.propagate()

            self.__prove_conflicting_conditions()
            self.__prove_reachability()
            self.__prove_contradictions()
            self.__prove_coverage()
            self.__prove_overlays()
            self.__prove_field_positions()
            self.__prove_message_size()

            self.error.propagate()

    def copy(
        self,
        identifier: StrID = None,
        structure: Sequence[Link] = None,
        types: Mapping[Field, mty.Type] = None,
        aspects: Mapping[ID, Mapping[ID, Sequence[expr.Expr]]] = None,
        location: Location = None,
        error: RecordFluxError = None,
    ) -> "Message":
        return Message(
            identifier if identifier else self.identifier,
            structure if structure else copy(self.structure),
            types if types else copy(self.types),
            aspects if aspects else copy(self.aspects),
            location if location else self.location,
            error if error else self.error,
        )

    def proven(self, skip_proof: bool = False) -> "Message":
        return copy(self)

    def is_possibly_empty(self, field: Field) -> bool:
        if isinstance(self.types[field], mty.Scalar):
            return False

        for p in self.paths(FINAL):
            if not any(l.target == field for l in p):
                continue
            empty_field = expr.Equal(expr.Size(field.name), expr.Number(0))
            proof = self.__prove_path_property(empty_field, p)
            if proof.result == expr.ProofResult.SAT:
                return True

        return False

    def set_refinements(self, refinements: List["Refinement"]) -> None:
        if any(r.pdu != self for r in refinements):
            fatal_fail("setting refinements for different message", Subsystem.MODEL)
        self._refinements = refinements

    @property
    def type_(self) -> rty.Message:
        return rty.Message(
            self.full_name,
            {tuple(l.target.name for l in p if l.target != FINAL) for p in self.paths(FINAL)}
            if self.structure
            else set(),
            {f.identifier: t.type_ for f, t in self.types.items()},
            [rty.Refinement(r.field.identifier, r.sdu.type_, r.package) for r in self._refinements],
            self.is_definite,
        )

    @property
    def has_fixed_size(self) -> bool:
        return len(self.paths(FINAL)) <= 1 and not (
            {v.identifier for l in self.structure for v in l.size.variables()}
            - set(self._type_literals.keys())
        )

    @property
    def has_bounded_size(self) -> bool:
        """
        Return true if the message has a bounded size.

        All messages containing a reference to `Message` in a size aspect are considered unbounded,
        even if the affected field size is restricted by a condition.
        """
        return all(
            v.identifier != ID("Message") for l in self.structure for v in l.size.variables()
        )

    @property
    def is_definite(self) -> bool:
        """
        Return true if the message has a bounded size and no optional fields.

        Messages with a First or Last attribute in a size aspect are not yet supported and
        therefore considered as not definite.
        """
        return (
            len(self.paths(FINAL)) <= 1
            and self.has_bounded_size
            and all(
                not l.size.findall(lambda x: isinstance(x, (expr.First, expr.Last)))
                for l in self.structure
                for v in l.size.variables()
            )
        )

    def size(self, field_values: Mapping[Field, expr.Expr] = None) -> expr.Expr:
        field_values = field_values if field_values else {}

        def remove_variable_prefix(expression: expr.Expr) -> expr.Expr:
            if isinstance(expression, expr.Variable) and expression.name.startswith("RFLX_"):
                return expr.Variable(
                    expression.name[5:],
                    expression.negative,
                    expression.immutable,
                    expression.type_,
                    location=expression.location,
                )
            return expression

        if not self.structure:
            return expr.Number(0)

        fields = set(field_values)
        values = [
            expr.Equal(expr.Variable(f.name), v, location=v.location)
            for f, v in field_values.items()
        ]
        aggregate_sizes = [
            expr.Equal(expr.Size(f.name), expr.Number(len(v.elements) * 8), location=v.location)
            for f, v in field_values.items()
            if isinstance(v, expr.Aggregate)
        ]
        composite_sizes = [
            expr.Equal(
                expr.Size(f.name),
                expr.Size(
                    expr.Variable(
                        "RFLX_" + v.identifier,
                        type_=v.type_,
                        location=v.location,
                    )
                ),
            )
            for f, v in field_values.items()
            if isinstance(self.types[f], mty.Composite) and isinstance(v, expr.Variable)
        ]
        failures = []

        for path in self.paths(FINAL):
            if not self.has_fixed_size and fields != set(
                l.target for l in path if l.target != FINAL
            ):
                continue
            message_size = expr.Add(
                *[
                    expr.Size(link.target.name)
                    for link in path
                    if link.target != FINAL and link.first == expr.UNDEFINED
                ]
            )
            link_expressions = [fact for link in path for fact in self.__link_expression(link)]
            proof = expr.Equal(expr.Size("Message"), message_size).check(
                [
                    *aggregate_sizes,
                    *composite_sizes,
                    *link_expressions,
                    *values,
                    *self.type_constraints(expr.TRUE),
                ]
            )

            if proof.result == expr.ProofResult.SAT:
                return (
                    message_size.substituted(mapping=to_mapping(aggregate_sizes + composite_sizes))
                    .substituted(mapping=to_mapping(link_expressions))
                    .substituted(mapping=to_mapping(values))
                    .substituted(mapping=to_mapping(self.type_constraints(expr.TRUE)))
                    .substituted(remove_variable_prefix)
                    .simplified()
                )

            failures.append((path, proof.error))

        error = RecordFluxError()
        error.append(
            f"unable to calculate size for message \"{self.identifier}'("
            + ", ".join(f"{f.identifier} => {v}" for f, v in field_values.items())
            + ')"',
            Subsystem.MODEL,
            Severity.ERROR,
            self.location,
        )
        for path, proof_error in failures:
            error.append(
                "on path " + " -> ".join([l.target.name for l in path]),
                Subsystem.MODEL,
                Severity.INFO,
                self.location,
            )
            error.extend(
                [
                    (f'unsatisfied "{m}"', Subsystem.MODEL, Severity.INFO, locn)
                    for m, locn in proof_error
                ]
            )
        error.propagate()

        assert False

    def max_size(self) -> expr.Number:
        if not self.structure:
            return expr.Number(0)

        if not self.has_bounded_size:
            fail(
                "unable to calculate maximum size of unbounded message",
                Subsystem.MODEL,
                Severity.ERROR,
                self.location,
            )

        max_size = expr.Number(0)

        for path in self.paths(FINAL):
            max_size = max(max_size, self.__max_value(expr.Size("Message"), path))

        return max_size

    def max_field_sizes(self) -> Dict[Field, expr.Number]:
        if not self.structure:
            return {}

        if not self.has_bounded_size:
            fail(
                "unable to calculate maximum field sizes of unbounded message",
                Subsystem.MODEL,
                Severity.ERROR,
                self.location,
            )

        result = {f: expr.Number(0) for f in self.fields}

        for path in self.paths(FINAL):
            for l in path[:-1]:
                result[l.target] = max(
                    result[l.target], self.__max_value(expr.Size(l.target.name), path)
                )

        return result

    def __max_value(self, target: expr.Expr, path: Tuple[Link, ...]) -> expr.Number:
        message_size = expr.Add(
            *[
                expr.Size(link.target.name)
                for link in path
                if link.target != FINAL and link.first == expr.UNDEFINED
            ]
        )
        link_expressions = [fact for link in path for fact in self.__link_expression(link)]
        return expr.max_value(
            target,
            [
                expr.Equal(expr.Size("Message"), message_size),
                *link_expressions,
                *self.type_constraints(expr.TRUE),
            ],
        )

    def __verify_expression_types(self) -> None:
        types: Dict[ID, mty.Type] = {}

        def typed_variable(expression: expr.Expr) -> expr.Expr:
            expression = copy(expression)
            if isinstance(expression, expr.Variable):
                if expression.name.lower() == "message":
                    expression.type_ = rty.OPAQUE
                elif expression.identifier in types:
                    expression.type_ = types[expression.identifier].type_
                elif expression.identifier in self._qualified_enum_literals:
                    expression.type_ = self._qualified_enum_literals[expression.identifier].type_
                elif expression.identifier in self._type_literals:
                    expression.type_ = self._type_literals[expression.identifier].type_
            return expression

        for p in self.paths(FINAL):
            types = {}
            path = []
            for l in p:
                try:
                    # check for contradictions in conditions of path
                    proof = self.__prove_path_property(expr.TRUE, p)
                    if proof.result == expr.ProofResult.UNSAT:
                        break
                except expr.Z3TypeError:
                    pass

                path.append(l.target)

                if l.source in self.types:
                    types[l.source.identifier] = self.types[l.source]

                for expression in [
                    l.condition.substituted(typed_variable),
                    l.size.substituted(typed_variable),
                    l.first.substituted(typed_variable),
                ]:
                    if expression == expr.UNDEFINED:
                        continue

                    error = expression.check_type(rty.Any())

                    self.error.extend(error)

                    if error.check():
                        self.error.append(
                            "on path " + " -> ".join(f.name for f in path),
                            Subsystem.MODEL,
                            Severity.INFO,
                            expression.location,
                        )

    def __verify_expressions(self) -> None:
        for f in (INITIAL, *self.fields):
            for l in self.outgoing(f):
                self.__check_attributes(l.condition, l.condition.location)
                self.__check_first_expression(l, l.first.location)
                self.__check_size_expression(l)

    def __check_attributes(self, expression: expr.Expr, location: Location = None) -> None:
        for a in expression.findall(lambda x: isinstance(x, expr.Attribute)):
            if isinstance(a, expr.Size) and not (
                isinstance(a.prefix, expr.Variable)
                and (
                    a.prefix.name == "Message"
                    or Field(a.prefix.name) in self.fields
                    or a.prefix.identifier in self._type_literals
                )
            ):
                self.error.append(
                    f'invalid use of size attribute for "{a.prefix}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    location,
                )

    def __check_first_expression(self, link: Link, location: Location = None) -> None:
        if link.first != expr.UNDEFINED and not isinstance(link.first, expr.First):
            self.error.append(
                f'invalid First for field "{link.target.name}"',
                Subsystem.MODEL,
                Severity.ERROR,
                location,
            )

    def __check_size_expression(self, link: Link) -> None:
        if link.target == FINAL and link.size != expr.UNDEFINED:
            self.error.append(
                f'size attribute for final field in "{self.identifier}"',
                Subsystem.MODEL,
                Severity.ERROR,
                link.size.location,
            )
        if link.target != FINAL and link.target in self.types:
            t = self.types[link.target]
            unconstrained = isinstance(t, (mty.Opaque, mty.Sequence))
            if not unconstrained and link.size != expr.UNDEFINED:
                self.error.append(
                    f'fixed size field "{link.target.name}" with size aspect',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    link.target.identifier.location,
                )
            if unconstrained and link.size == expr.UNDEFINED:
                self.error.append(
                    f'unconstrained field "{link.target.name}" without size aspect',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    link.target.identifier.location,
                )

    def __verify_checksums(self) -> None:
        def valid_lower(expression: expr.Expr) -> bool:
            return isinstance(expression, expr.First) or (
                isinstance(expression, expr.Add)
                and len(expression.terms) == 2
                and isinstance(expression.terms[0], expr.Last)
                and expression.terms[1] == expr.Number(1)
            )

        def valid_upper(expression: expr.Expr) -> bool:
            return isinstance(expression, expr.Last) or (
                isinstance(expression, expr.Sub)
                and isinstance(expression.left, expr.First)
                and expression.right == expr.Number(1)
            )

        for name, expressions in self.checksums.items():  # pylint: disable=too-many-nested-blocks
            if Field(name) not in self.fields:
                self.error.append(
                    f'checksum definition for unknown field "{name}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    name.location,
                )

            for e in expressions:
                if not (
                    isinstance(e, (expr.Variable, expr.Size))
                    or (
                        isinstance(e, expr.ValueRange)
                        and valid_lower(e.lower)
                        and valid_upper(e.upper)
                    )
                ):
                    self.error.append(
                        f'unsupported expression "{e}" in definition of checksum "{name}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        e.location,
                    )

                for v in e.findall(lambda x: isinstance(x, expr.Variable)):
                    assert isinstance(v, expr.Variable)

                    if Field(v.name) not in self.fields:
                        self.error.append(
                            f'unknown field "{v.name}" referenced'
                            f' in definition of checksum "{name}"',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            v.location,
                        )

                if isinstance(e, expr.ValueRange):
                    lower = e.lower.findall(lambda x: isinstance(x, expr.Variable))[0]
                    upper = e.upper.findall(lambda x: isinstance(x, expr.Variable))[0]

                    assert isinstance(lower, expr.Variable)
                    assert isinstance(upper, expr.Variable)

                    if lower != upper:
                        upper_field = (
                            Field(upper.name) if upper.name.lower() != "message" else FINAL
                        )
                        lower_field = (
                            Field(lower.name) if lower.name.lower() != "message" else INITIAL
                        )
                        for p in self.paths(upper_field):
                            if not any(lower_field == l.source for l in p):
                                self.error.append(
                                    f'invalid range "{e}" in definition of checksum "{name}"',
                                    Subsystem.MODEL,
                                    Severity.ERROR,
                                    e.location,
                                )

        checked = {
            e.prefix.identifier
            for e in self.field_condition(FINAL).findall(
                lambda x: isinstance(x, expr.ValidChecksum)
            )
            if isinstance(e, expr.ValidChecksum) and isinstance(e.prefix, expr.Variable)
        }
        for name in set(self.checksums) - checked:
            self.error.append(
                f'no validity check of checksum "{name}"',
                Subsystem.MODEL,
                Severity.ERROR,
                name.location,
            )
        for name in checked - set(self.checksums):
            self.error.append(
                f'validity check for undefined checksum "{name}"',
                Subsystem.MODEL,
                Severity.ERROR,
                name.location,
            )

    def __prove_conflicting_conditions(self) -> None:
        for f in (INITIAL, *self.fields):
            for i1, c1 in enumerate(self.outgoing(f)):
                for i2, c2 in enumerate(self.outgoing(f)):
                    if i1 < i2:
                        conflict = expr.And(c1.condition, c2.condition)
                        proof = conflict.check(self.type_constraints(conflict))
                        if proof.result == expr.ProofResult.SAT:
                            c1_message = str(c1.condition).replace("\n", " ")
                            c2_message = str(c2.condition).replace("\n", " ")
                            self.error.append(
                                f'conflicting conditions for field "{f.name}"',
                                Subsystem.MODEL,
                                Severity.ERROR,
                                f.identifier.location,
                            )
                            self.error.append(
                                f"condition {i1} ({f.identifier} -> {c1.target.identifier}):"
                                f" {c1_message}",
                                Subsystem.MODEL,
                                Severity.INFO,
                                c1.condition.location,
                            )
                            self.error.append(
                                f"condition {i2} ({f.identifier} -> {c2.target.identifier}):"
                                f" {c2_message}",
                                Subsystem.MODEL,
                                Severity.INFO,
                                c2.condition.location,
                            )

    def __prove_reachability(self) -> None:
        def has_final(field: Field) -> bool:
            if field == FINAL:
                return True
            for o in self.outgoing(field):
                if has_final(o.target):
                    return True
            return False

        for f in (INITIAL, *self.fields):
            if not has_final(f):
                self.error.append(
                    f'no path to FINAL for field "{f.name}" in "{self.identifier}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    f.identifier.location,
                )

        for f in (*self.fields, FINAL):
            paths = []
            for path in self.paths(f):
                facts = [fact for link in path for fact in self.__link_expression(link)]
                last_field = path[-1].target
                outgoing = self.outgoing(last_field)
                if last_field != FINAL and outgoing:
                    facts.append(
                        expr.Or(
                            *[o.condition for o in outgoing],
                            location=last_field.identifier.location,
                        )
                    )
                proof = expr.TRUE.check(facts)
                if proof.result == expr.ProofResult.SAT:
                    break

                paths.append((path, proof.error))
            else:
                self.error.append(
                    f'unreachable field "{f.name}" in "{self.identifier}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    f.identifier.location,
                )
                for index, (path, errors) in enumerate(sorted(paths)):
                    self.error.append(
                        f"path {index} (" + " -> ".join([l.target.name for l in path]) + "):",
                        Subsystem.MODEL,
                        Severity.INFO,
                        f.identifier.location,
                    )
                    self.error.extend(
                        [
                            (f'unsatisfied "{m}"', Subsystem.MODEL, Severity.INFO, l)
                            for m, l in errors
                        ]
                    )

    def __prove_contradictions(self) -> None:
        for f in (INITIAL, *self.fields):
            contradictions = []
            paths = 0
            for path in self.paths(f):
                facts = [fact for link in path for fact in self.__link_expression(link)]
                for c in self.outgoing(f):
                    paths += 1
                    contradiction = c.condition
                    constraints = self.message_constraints() + self.type_constraints(contradiction)
                    proof = contradiction.check([*constraints, *facts])
                    if proof.result == expr.ProofResult.SAT:
                        continue

                    contradictions.append((path, c.condition, proof.error))

            if paths == len(contradictions):
                for path, cond, errors in sorted(contradictions):
                    self.error.append(
                        f'contradicting condition in "{self.identifier}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        cond.location,
                    )
                    self.error.extend(
                        [
                            (
                                f'on path: "{l.target.identifier}"',
                                Subsystem.MODEL,
                                Severity.INFO,
                                l.target.identifier.location,
                            )
                            for l in path
                        ]
                    )
                    self.error.extend(
                        [
                            (f'unsatisfied "{m}"', Subsystem.MODEL, Severity.INFO, l)
                            for m, l in errors
                        ]
                    )

    def __prove_coverage(self) -> None:
        """
        Prove that the fields of a message cover all message bits, i.e. there are no holes in the
        message definition.

        Idea: Let f be the bits covered by the message. By definition
            (1) f >= Message'First and f <= Message'Last
        holds. For every field add a conjunction of the form
            (2) Not(f >= Field'First and f <= Field'Last),
        effectively pruning the range that this field covers from the bit range of the message. For
        the overall expression, prove that it is false for all f, i.e. no bits are left.
        """
        for path in [p[:-1] for p in self.paths(FINAL) if p]:

            facts: Sequence[expr.Expr]

            # Calculate (1)
            facts = [
                expr.GreaterEqual(expr.Variable("f"), expr.First("Message")),
                expr.LessEqual(expr.Variable("f"), expr.Last("Message")),
            ]
            # Calculate (2) for all fields
            facts.extend(
                [
                    expr.Not(
                        expr.And(
                            expr.GreaterEqual(expr.Variable("f"), self.__target_first(l)),
                            expr.LessEqual(expr.Variable("f"), self.__target_last(l)),
                            location=l.location,
                        )
                    )
                    for l in path
                ]
            )

            # Define that the end of the last field of a path is the end of the message
            facts.append(
                expr.Equal(self.__target_last(path[-1]), expr.Last("Message"), self.location)
            )

            # Constraints for links and types
            facts.extend([f for l in path for f in self.__link_expression(l)])

            # Coverage expression must be False, i.e. no bits left
            proof = expr.TRUE.check(facts)
            if proof.result == expr.ProofResult.SAT:
                self.error.append(
                    "path does not cover whole message",
                    Subsystem.MODEL,
                    Severity.ERROR,
                    self.identifier.location,
                )
                self.error.extend(
                    [
                        (
                            f'on path: "{l.target.identifier}"',
                            Subsystem.MODEL,
                            Severity.INFO,
                            l.target.identifier.location,
                        )
                        for l in path
                    ]
                )
                return

    def __prove_overlays(self) -> None:
        for f in (INITIAL, *self.fields):
            for p, l in [(p, p[-1]) for p in self.paths(f) if p]:
                if l.first != expr.UNDEFINED and isinstance(l.first, expr.First):
                    facts = [f for l in p for f in self.__link_expression(l)]
                    overlaid = expr.Equal(
                        self.__target_last(l), expr.Last(l.first.prefix), l.location
                    )
                    proof = overlaid.check(facts)
                    if proof.result != expr.ProofResult.SAT:
                        self.error.append(
                            f'field "{f.name}" not congruent with'
                            f' overlaid field "{l.first.prefix}"',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            self.identifier.location,
                        )
                        self.error.extend(
                            [
                                (f'unsatisfied "{m}"', Subsystem.MODEL, Severity.INFO, l)
                                for m, l in proof.error
                            ]
                        )

    def __prove_field_positions(self) -> None:
        for f in (*self.fields, FINAL):
            for path in self.paths(f):

                last = path[-1]
                negative = expr.Less(self.__target_size(last), expr.Number(0), last.size.location)
                start = expr.GreaterEqual(
                    self.__target_first(last), expr.First("Message"), last.location
                )

                facts = [fact for link in path for fact in self.__link_expression(link)]

                outgoing = self.outgoing(f)
                if f != FINAL and outgoing:
                    facts.append(
                        expr.Or(*[o.condition for o in outgoing], location=f.identifier.location)
                    )

                facts.extend(self.type_constraints(negative))
                facts.extend(self.type_constraints(start))

                proof = expr.TRUE.check(facts)

                # Only check positions of reachable paths
                if proof.result != expr.ProofResult.SAT:
                    continue

                proof = negative.check(facts)
                if proof.result != expr.ProofResult.UNSAT:
                    path_message = " -> ".join([l.target.name for l in path])
                    self.error.append(
                        f'negative size for field "{f.name}" ({path_message})',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        f.identifier.location,
                    )
                    return

                proof = start.check(facts)
                if proof.result != expr.ProofResult.SAT:
                    path_message = " -> ".join([last.target.name for last in path])
                    self.error.append(
                        f'negative start for field "{f.name}" ({path_message})',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.identifier.location,
                    )
                    self.error.extend(
                        [
                            (f'unsatisfied "{m}"', Subsystem.MODEL, Severity.INFO, locn)
                            for m, locn in proof.error
                        ]
                    )
                    return

                if f in self.types:
                    t = self.types[f]
                    if not isinstance(t, mty.Opaque):
                        continue
                    element_size = t.element_size
                    start_aligned = expr.Not(
                        expr.Equal(
                            expr.Mod(self.__target_first(last), element_size),
                            expr.Number(1),
                            last.location,
                        )
                    )
                    proof = start_aligned.check(
                        [
                            *facts,
                            *self.message_constraints(),
                            *self.type_constraints(start_aligned),
                        ]
                    )
                    if proof.result != expr.ProofResult.UNSAT:
                        path_message = " -> ".join([p.target.name for p in path])
                        self.error.append(
                            f'opaque field "{f.name}" not aligned to {element_size} bit boundary'
                            f" ({path_message})",
                            Subsystem.MODEL,
                            Severity.ERROR,
                            f.identifier.location,
                        )
                        return

                    is_multiple_of_element_size = expr.Not(
                        expr.Equal(
                            expr.Mod(self.__target_size(last), element_size),
                            expr.Number(0),
                            last.location,
                        )
                    )
                    proof = is_multiple_of_element_size.check(
                        [
                            *facts,
                            *self.message_constraints(),
                            *self.type_constraints(is_multiple_of_element_size),
                        ]
                    )
                    if proof.result != expr.ProofResult.UNSAT:
                        path_message = " -> ".join([p.target.name for p in path])
                        self.error.append(
                            f'size of opaque field "{f.name}" not multiple of {element_size} bit'
                            f" ({path_message})",
                            Subsystem.MODEL,
                            Severity.ERROR,
                            f.identifier.location,
                        )
                        return

    def __prove_message_size(self) -> None:
        """
        Prove that all message paths lead to a message with a size that is a multiple of 8 bit.
        """
        type_constraints = self.type_constraints(expr.TRUE)
        field_size_constraints = [
            expr.Equal(expr.Mod(expr.Size(f.name), expr.Number(8)), expr.Number(0))
            for f, t in self.types.items()
            if isinstance(t, (mty.Opaque, mty.Sequence))
        ]

        for path in [p[:-1] for p in self.paths(FINAL) if p]:
            message_size = expr.Add(
                *[
                    expr.Size(link.target.name)
                    for link in path
                    if link.target != FINAL and link.first == expr.UNDEFINED
                ]
            )
            facts = [
                *[fact for link in path for fact in self.__link_expression(link)],
                *type_constraints,
                *field_size_constraints,
            ]
            proof = expr.NotEqual(expr.Mod(message_size, expr.Number(8)), expr.Number(0)).check(
                facts
            )
            if proof.result == expr.ProofResult.SAT:
                self.error.append(
                    "message size must be multiple of 8 bit",
                    Subsystem.MODEL,
                    Severity.ERROR,
                    self.identifier.location,
                )
                self.error.append(
                    "on path " + " -> ".join(l.target.name for l in path),
                    Subsystem.MODEL,
                    Severity.INFO,
                    self.identifier.location,
                )
                return

    def __prove_path_property(self, prop: expr.Expr, path: Sequence[Link]) -> expr.Proof:
        conditions = [l.condition for l in path if l.condition != expr.TRUE]
        sizes = [
            expr.Equal(expr.Size(l.target.name), l.size) for l in path if l.size != expr.UNDEFINED
        ]
        return prop.check([*self.type_constraints(prop), *conditions, *sizes])

    @staticmethod
    def __target_first(link: Link) -> expr.Expr:
        if link.source == INITIAL:
            return expr.First("Message")
        if link.first != expr.UNDEFINED:
            return link.first
        return expr.Add(expr.Last(link.source.name), expr.Number(1), location=link.location)

    def __target_size(self, link: Link) -> expr.Expr:
        if link.size != expr.UNDEFINED:
            return link.size
        return self.field_size(link.target)

    def __target_last(self, link: Link) -> expr.Expr:
        return expr.Sub(
            expr.Add(self.__target_first(link), self.__target_size(link)),
            expr.Number(1),
            link.target.identifier.location,
        )

    def __link_expression(self, link: Link) -> List[expr.Expr]:
        name = link.target.name
        target_first = self.__target_first(link)
        target_size = self.__target_size(link)
        target_last = self.__target_last(link)
        return [
            expr.Equal(expr.First(name), target_first, target_first.location or self.location),
            *(
                [
                    expr.Equal(expr.Size(name), target_size, target_size.location or self.location),
                    expr.Equal(expr.Last(name), target_last, target_last.location or self.location),
                ]
                if target_size != expr.UNDEFINED
                else []
            ),
            expr.GreaterEqual(expr.First("Message"), expr.Number(0), self.location),
            expr.GreaterEqual(expr.Last("Message"), expr.Last(name), self.location),
            expr.GreaterEqual(expr.Last("Message"), expr.First("Message"), self.location),
            expr.Equal(
                expr.Size("Message"),
                expr.Add(expr.Sub(expr.Last("Message"), expr.First("Message")), expr.Number(1)),
                self.location,
            ),
            *expression_list(link.condition),
        ]


class DerivedMessage(Message):
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        identifier: StrID,
        base: Message,
        structure: Sequence[Link] = None,
        types: Mapping[Field, mty.Type] = None,
        aspects: Mapping[ID, Mapping[ID, Sequence[expr.Expr]]] = None,
        location: Location = None,
        error: RecordFluxError = None,
    ) -> None:
        super().__init__(
            identifier,
            structure if structure else copy(base.structure),
            types if types else copy(base.types),
            aspects if aspects else copy(base.aspects),
            location if location else base.location,
            error if error else base.error,
        )
        self.base = base

    def copy(
        self,
        identifier: StrID = None,
        structure: Sequence[Link] = None,
        types: Mapping[Field, mty.Type] = None,
        aspects: Mapping[ID, Mapping[ID, Sequence[expr.Expr]]] = None,
        location: Location = None,
        error: RecordFluxError = None,
    ) -> "DerivedMessage":
        return DerivedMessage(
            identifier if identifier else self.identifier,
            self.base,
            structure if structure else copy(self.structure),
            types if types else copy(self.types),
            aspects if aspects else copy(self.aspects),
            location if location else self.location,
            error if error else self.error,
        )

    def proven(self, skip_proof: bool = False) -> "DerivedMessage":
        return copy(self)


class UnprovenMessage(AbstractMessage):
    # pylint: disable=too-many-arguments
    def copy(
        self,
        identifier: StrID = None,
        structure: Sequence[Link] = None,
        types: Mapping[Field, mty.Type] = None,
        aspects: Mapping[ID, Mapping[ID, Sequence[expr.Expr]]] = None,
        location: Location = None,
        error: RecordFluxError = None,
    ) -> "UnprovenMessage":
        return UnprovenMessage(
            identifier if identifier else self.identifier,
            structure if structure else copy(self.structure),
            types if types else copy(self.types),
            aspects if aspects else copy(self.aspects),
            location if location else self.location,
            error if error else self.error,
        )

    def proven(self, skip_proof: bool = False) -> Message:
        return Message(
            identifier=self.identifier,
            structure=self.structure,
            types=self.types,
            location=self.location,
            error=self.error,
            state=self._state,
            skip_proof=skip_proof,
        )

    @ensure(lambda result: valid_message_field_types(result))
    def merged(self) -> "UnprovenMessage":
        def prune_dangling_fields(
            structure: List[Link], types: Dict[Field, mty.Type]
        ) -> Tuple[List[Link], Dict[Field, mty.Type]]:
            dangling = []
            progress = True
            while progress:
                progress = False
                fields = {x for l in structure for x in (l.source, l.target) if x != FINAL}
                for s in fields:
                    if all(l.source != s for l in structure):
                        dangling.append(s)
                        progress = True
                structure = [l for l in structure if l.target not in dangling]

            return (
                structure,
                {k: v for k, v in types.items() if k not in dangling},
            )

        message = self

        while True:
            inner_messages = [
                (f, t) for f, t in message.types.items() if isinstance(t, AbstractMessage)
            ]

            if not inner_messages:
                break

            field, inner_message = inner_messages.pop(0)
            inner_message = inner_message.prefixed(f"{field.name}_")

            name_conflicts = [
                f for f in message.fields for g in inner_message.fields if f.name == g.name
            ]

            if name_conflicts:
                conflicting = name_conflicts.pop(0)
                self.error.append(
                    f'name conflict for "{conflicting.identifier}" in "{message.identifier}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    conflicting.identifier.location,
                )
                self.error.append(
                    f'when merging message "{inner_message.identifier}"',
                    Subsystem.MODEL,
                    Severity.INFO,
                    inner_message.location,
                )
                self.error.append(
                    f'into field "{field.name}"',
                    Subsystem.MODEL,
                    Severity.INFO,
                    field.identifier.location,
                )

            structure = []

            for link in message.structure:
                if link.target == field:
                    initial_link = inner_message.outgoing(INITIAL)[0]
                    structure.append(
                        Link(
                            link.source,
                            initial_link.target,
                            link.condition,
                            initial_link.size,
                            link.first,
                            link.location,
                        )
                    )
                elif link.source == field:
                    for final_link in inner_message.incoming(FINAL):
                        merged_condition = expr.And(link.condition, final_link.condition)
                        proof = merged_condition.check(
                            [
                                *inner_message.message_constraints(),
                                *inner_message.type_constraints(merged_condition),
                                inner_message.field_condition(final_link.source),
                            ]
                        )
                        if proof.result != expr.ProofResult.UNSAT:
                            structure.append(
                                Link(
                                    final_link.source,
                                    link.target,
                                    merged_condition.simplified(),
                                    link.size,
                                    link.first,
                                    link.location,
                                )
                            )
                else:
                    structure.append(link)

            structure.extend(
                l for l in inner_message.structure if l.target != FINAL and l.source != INITIAL
            )

            types = {
                **{f: t for f, t in message.types.items() if f != field},
                **inner_message.types,
            }

            structure, types = prune_dangling_fields(structure, types)
            if not structure or not types:
                fail(
                    f'empty message type when merging field "{field.identifier}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    field.identifier.location,
                )

            message = message.copy(structure=structure, types=types)

        return message


class UnprovenDerivedMessage(UnprovenMessage):
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        identifier: StrID,
        base: Union[UnprovenMessage, Message],
        structure: Sequence[Link] = None,
        types: Mapping[Field, mty.Type] = None,
        aspects: Mapping[ID, Mapping[ID, Sequence[expr.Expr]]] = None,
        location: Location = None,
        error: RecordFluxError = None,
    ) -> None:
        super().__init__(
            identifier,
            structure if structure else copy(base.structure),
            types if types else copy(base.types),
            aspects if aspects else copy(base.aspects),
            location if location else base.location,
            error if error else base.error,
        )
        self.error.extend(base.error)
        self.base = base

        if isinstance(base, (UnprovenDerivedMessage, DerivedMessage)):
            self.error.append(
                f'illegal derivation "{self.identifier}"',
                Subsystem.MODEL,
                Severity.ERROR,
                self.location,
            )
            self.error.append(
                f'illegal base message type "{base.identifier}"',
                Subsystem.MODEL,
                Severity.INFO,
                base.location,
            )
            self.error.propagate()

    def copy(
        self,
        identifier: StrID = None,
        structure: Sequence[Link] = None,
        types: Mapping[Field, mty.Type] = None,
        aspects: Mapping[ID, Mapping[ID, Sequence[expr.Expr]]] = None,
        location: Location = None,
        error: RecordFluxError = None,
    ) -> "UnprovenDerivedMessage":
        return UnprovenDerivedMessage(
            identifier if identifier else self.identifier,
            self.base,
            structure if structure else copy(self.structure),
            types if types else copy(self.types),
            aspects if aspects else copy(self.aspects),
            location if location else self.location,
            error if error else self.error,
        )

    def proven(self, skip_proof: bool = False) -> DerivedMessage:
        return DerivedMessage(
            self.identifier,
            self.base if isinstance(self.base, Message) else self.base.proven(),
            self.structure,
            self.types,
            self.aspects,
            self.location,
            self.error,
        )


class Refinement(mty.Type):
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        package: StrID,
        pdu: Message,
        field: Field,
        sdu: Message,
        condition: expr.Expr = expr.TRUE,
        location: Location = None,
        error: RecordFluxError = None,
    ) -> None:
        package = ID(package)

        super().__init__(
            package * "__REFINEMENT__"
            f"{sdu.identifier.flat}__{pdu.identifier.flat}__{field.name}__",
            location,
            error,
        )

        self.pdu = pdu
        self.field = field
        self.sdu = sdu
        self.condition = condition

        self.error = error or RecordFluxError()
        if len(package.parts) != 1:
            self.error.append(
                f'unexpected format of package name "{package}"',
                Subsystem.MODEL,
                Severity.ERROR,
                package.location,
            )

        for f, t in pdu.types.items():
            if f == field:
                if not isinstance(t, mty.Opaque):
                    self.error.append(
                        f'invalid type of field "{field.name}" in refinement of "{pdu.identifier}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        field.identifier.location,
                    )
                    self.error.append(
                        "expected field of type Opaque",
                        Subsystem.MODEL,
                        Severity.INFO,
                        f.identifier.location,
                    )
                break
        else:
            self.error.append(
                f'invalid field "{field.name}" in refinement of "{pdu.identifier}"',
                Subsystem.MODEL,
                Severity.ERROR,
                field.identifier.location,
            )

        for variable in condition.variables():
            literals = mty.enum_literals(pdu.types.values(), self.package)
            if Field(str(variable.name)) not in pdu.fields and variable.identifier not in literals:
                self.error.append(
                    f'unknown field or literal "{variable.identifier}" in refinement'
                    f' condition of "{pdu.identifier}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    variable.location,
                )

    def __str__(self) -> str:
        condition = f"\n   if {self.condition}" if self.condition != expr.TRUE else ""
        return f"for {self.pdu.name} use ({self.field.name} => {self.sdu.name}){condition}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.package == other.package
                and self.pdu == other.pdu
                and self.field == other.field
                and self.sdu == other.sdu
            )
        return NotImplemented


def expression_list(expression: expr.Expr) -> Sequence[expr.Expr]:
    if isinstance(expression, expr.And):
        return expression.terms
    return [expression]


def to_mapping(facts: Sequence[expr.Expr]) -> Dict[expr.Name, expr.Expr]:
    return {
        f.left: f.right
        for f in facts
        if isinstance(f, (expr.Equal, expr.GreaterEqual)) and isinstance(f.left, expr.Name)
    }



from abc import abstractmethod
from collections import deque
from enum import Enum, auto
from pathlib import Path
from typing import Deque, List, NoReturn, Optional, Tuple, TypeVar, Union

from rflx.common import Base, verbose_repr


class Location(Base):
    def __init__(
        self,
        start: Tuple[int, int],
        source: Path = None,
        end: Tuple[int, int] = None,
        verbose: bool = False,
    ):
        self.__source = source
        self.__start = start
        self.__end = end
        self.__verbose = verbose

    def __str__(self) -> str:
        def linecol_str(linecol: Tuple[int, int]) -> str:
            return f"{linecol[0]}:{linecol[1]}"

        start = f":{linecol_str(self.__start)}"
        end = f"-{linecol_str(self.__end)}" if self.__end and self.__verbose else ""
        return f"{self.__source if self.__source else '<stdin>'}{start}{end}"

    def __lt__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.start < other.start
        return NotImplemented

    @property
    def source(self) -> Optional[Path]:
        return self.__source

    @property
    def start(self) -> Tuple[int, int]:
        return self.__start

    @property
    def end(self) -> Optional[Tuple[int, int]]:
        return self.__end


class Subsystem(Enum):
    PARSER = auto()
    MODEL = auto()
    GENERATOR = auto()
    CLI = auto()
    GRAPH = auto()
    PYRFLX = auto()
    ID = auto()

    def __str__(self) -> str:
        return str.lower(self.name)


class Severity(Enum):
    NONE = 1
    INFO = 2
    WARNING = 3
    ERROR = 4

    def __str__(self) -> str:
        return str.lower(self.name)


Self = TypeVar("Self", bound="BaseError")


class BaseError(Exception, Base):
    class Entry(Base):
        def __init__(
            self, message: str, subsystem: Subsystem, severity: Severity, location: Location = None
        ):
            self.__message = message
            self.__subsystem = subsystem
            self.__severity = severity
            self.__location = location

        @property
        def message(self) -> str:
            return self.__message

        @property
        def subsystem(self) -> Subsystem:
            return self.__subsystem

        @property
        def severity(self) -> Severity:
            return self.__severity

        @property
        def location(self) -> Optional[Location]:
            return self.__location

    @abstractmethod
    def __init__(self) -> None:
        super().__init__()
        self.__errors: Deque[BaseError.Entry] = deque()

    def __repr__(self) -> str:
        return verbose_repr(self, ["errors"])

    def __str__(self) -> str:
        def locn(entry: BaseError.Entry) -> str:
            if entry.location:
                return f"{entry.location}: "
            return ""

        return "\n".join(
            f"{locn(e)}{e.subsystem}: {e.severity}: {e.message}" for e in self.__errors
        )

    def __add__(self: Self, other: object) -> Self:
        if isinstance(other, BaseError):
            error = self.__class__()
            error.extend(self)
            error.extend(other)
            return error
        return NotImplemented

    def __iadd__(self: Self, other: object) -> Self:
        if isinstance(other, BaseError):
            self.extend(other)
            return self
        return NotImplemented

    @property
    def errors(self) -> Deque["BaseError.Entry"]:
        return self.__errors

    def append(
        self, message: str, subsystem: Subsystem, severity: Severity, location: Location = None
    ) -> None:
        self.__errors.append(BaseError.Entry(message, subsystem, severity, location))

    def appendleft(
        self, message: str, subsystem: Subsystem, severity: Severity, location: Location = None
    ) -> None:
        self.__errors.appendleft(BaseError.Entry(message, subsystem, severity, location))

    def extend(
        self,
        entries: Union[List[Tuple[str, Subsystem, Severity, Optional[Location]]], "BaseError"],
    ) -> None:
        if isinstance(entries, BaseError):
            self.__errors.extend(entries.errors)
        else:
            for message, subsystem, severity, location in entries:
                self.__errors.append(BaseError.Entry(message, subsystem, severity, location))

    def check(self) -> bool:
        return len(self.__errors) > 0

    def propagate(self) -> None:
        if self.check():
            raise self


class RecordFluxError(BaseError):
    """Error indicating an issue in an input or a known limitation."""

    def __init__(self) -> None:  # pylint: disable = useless-super-delegation
        super().__init__()


class FatalError(BaseError):
    """
    Error indicating a bug.

    This exception should never be caught outside of RecordFlux.
    """

    def __init__(self) -> None:  # pylint: disable = useless-super-delegation
        super().__init__()


def fail(
    message: str,
    subsystem: Subsystem,
    severity: Severity = Severity.ERROR,
    location: Location = None,
) -> NoReturn:
    _fail(RecordFluxError(), message, subsystem, severity, location)


def fatal_fail(
    message: str,
    subsystem: Subsystem,
    severity: Severity = Severity.ERROR,
    location: Location = None,
) -> NoReturn:
    _fail(FatalError(), message, subsystem, severity, location)


def _fail(
    error: BaseError,
    message: str,
    subsystem: Subsystem,
    severity: Severity = Severity.ERROR,
    location: Location = None,
) -> NoReturn:
    error.append(message, subsystem, severity, location)
    error.propagate()
    assert False


def warn(
    message: str,
    subsystem: Subsystem,
    severity: Severity = Severity.WARNING,
    location: Location = None,
) -> None:
    e = RecordFluxError()
    e.append(message, subsystem, severity, location)
    print(e)
